import os
import subprocess
import logging

from ..utils import working_directory_keeper
from .base import BaseRepo

logger = logging.getLogger(__name__)


class SvnCheckout(BaseRepo):

    vcs_control_dir = '.svn'

    def get_update(self, revision):
        """Ensure that target_dir is a branch of url at specified revision.

        If target_dir already exists, does a simple pull.
        Offline-mode: no branch nor pull, but update.
        """
        target_dir = self.target_dir
        url = self.url
        offline = self.offline

        rev_str = revision and '-r ' + revision or ''

        with working_directory_keeper:
            if not os.path.exists(target_dir):
                # TODO case of local url ?
                if offline:
                    raise IOError(
                        "svn checkout %s does not exist; cannot checkout "
                        "from %s (offline mode)" % (target_dir, url))

                os.chdir(os.path.split(target_dir)[0])
                logger.info("Checkouting %s ...", url)
                subprocess.check_call('svn checkout %s %s %s' % (
                    rev_str, url, target_dir), shell=True)
            else:
                os.chdir(target_dir)
                # TODO what if remote repo is actually local fs ?
                if offline:
                    logger.warning(
                        "Offline mode: keeping checkout %s in its current rev",
                        target_dir)
                else:
                    logger.info("Updating %s to location %s, revision %s...",
                                target_dir, url, revision)
                    # switch is necessary in order to move in tags
                    # TODO support also change of svn root url
                    subprocess.check_call('svn switch %s' % url, shell=True)
                    subprocess.check_call('svn up %s' % rev_str,
                                          shell=True)



"""VCS tests: Git."""

import os
import subprocess
from ..testing import COMMIT_USER_EMAIL
from ..testing import COMMIT_USER_NAME
from ..testing import VcsTestCase
from ..git import GitRepo
from ...utils import working_directory_keeper, WorkingDirectoryKeeper
from ...utils import check_output


def git_write_commit(repo_dir, filename, contents, msg="Unit test commit"):
    """Write specified file with contents, commit and return commit SHA.

    ``filename`` is actually a relative path from ``repo_dir``.
    """

    with WorkingDirectoryKeeper():  # independent from the main instance
        os.chdir(repo_dir)
        # repo configuration is local by default
        # needs to be done just once, but I prefer to do it a few useless
        # times than to forget it, since it's easy to turn into a sporadic
        # test breakage on continuous integration builds.
        subprocess.call(['git', 'config', 'user.email', COMMIT_USER_EMAIL])
        subprocess.call(['git', 'config', 'user.name', COMMIT_USER_NAME])
        with open(filename, 'w') as f:
            f.write(contents)
        subprocess.call(['git', 'add', filename])
        subprocess.call(['git', 'commit', '-m', msg])
        return check_output(['git', 'rev-parse', '--verify', 'HEAD']).strip()


class GitBaseTestCase(VcsTestCase):
    """Common utilities for Git test cases."""

    def create_src(self):
        os.chdir(self.src_dir)
        subprocess.call(['git', 'init', 'src-branch'])
        self.src_repo = os.path.join(self.src_dir, 'src-branch')
        self.commit_1_sha = git_write_commit(self.src_repo, 'tracked',
                                             "first", msg="initial commit")
        self.commit_2_sha = git_write_commit(self.src_repo, 'tracked',
                                             "last", msg="last commit")


class GitTestCase(GitBaseTestCase):

    def test_clone(self):
        """Git clone."""
        target_dir = os.path.join(self.dst_dir, "My clone")
        GitRepo(target_dir, self.src_repo)('master')

        self.assertTrue(os.path.isdir(target_dir))
        with open(os.path.join(target_dir, 'tracked')) as f:
            self.assertEquals(f.read().strip(), 'last')

    def test_archive(self):
        """Git clone, then archive"""
        repo = GitRepo(os.path.join(self.dst_dir, "My clone"), self.src_repo)
        repo('master')

        archive_dir = os.path.join(self.dst_dir, "archive directory")
        repo.archive(archive_dir)
        with open(os.path.join(archive_dir, 'tracked')) as f:
            self.assertEquals(f.readlines()[0].strip(), 'last')

    def test_clean(self):
        target_dir = os.path.join(self.dst_dir, "My clone")
        repo = GitRepo(target_dir, self.src_repo)
        try:
            repo.clean()
        except:
            self.fail("clean() should not fail if "
                      "clone not already done")

        repo('master')

        dirty_dir = os.path.join(repo.target_dir, 'dirty')
        os.mkdir(dirty_dir)
        dirty_files = (os.path.join(repo.target_dir, 'untracked'),
                       os.path.join(dirty_dir, 'untracked2'))
        for path in dirty_files:
            with open(path, 'w') as f:
                f.write('content')
        repo.clean()
        for path in dirty_files:
            self.failIf(os.path.exists(path),
                        "Untracked file should have been removed")
        self.failIf(os.path.exists(dirty_dir),
                    "Untracked dir should have been removed")

    def test_clone_on_sha(self):
        """Git clone."""
        target_dir = os.path.join(self.dst_dir, "My clone")
        repo = GitRepo(target_dir, self.src_repo)
        repo('master')
        sha = repo.parents()[0]
        target_dir = os.path.join(self.dst_dir, "My clone 2")
        repo = GitRepo(target_dir, self.src_repo)
        repo(sha)
        sha2 = repo.parents()[0]
        self.assertEqual(sha, sha2, 'Bad clone on SHA')

        # next call of get_update()
        repo(sha)

    def test_clone_on_sha_update(self):
        """Git clone."""
        target_dir = os.path.join(self.dst_dir, "My clone")
        repo = GitRepo(target_dir, self.src_repo)
        repo(self.commit_1_sha)
        self.assertEqual(repo.parents(), [self.commit_1_sha])
        repo(self.commit_2_sha)
        self.assertEqual(repo.parents(), [self.commit_2_sha])

        # new commit in origin will need to be fetched
        new_sha = git_write_commit(self.src_repo, 'tracked',
                                   "new contents", msg="new commit")
        repo(new_sha)
        self.assertEqual(repo.parents(), [new_sha])

        # see launchpad #1215873
        repo(new_sha)

    def test_uncommitted_changes(self):
        """GitRepo can detect uncommitted changes."""
        # initial cloning
        target_dir = os.path.join(self.dst_dir, "clone to update")
        repo = GitRepo(target_dir, self.src_repo)
        repo('master')

        self.assertFalse(repo.uncommitted_changes())

        # now with a local modification
        with open(os.path.join(target_dir, 'tracked'), 'w') as f:
            f.write('mod')
        self.assertTrue(repo.uncommitted_changes())


class GitBranchTestCase(GitBaseTestCase):

    def create_src(self):
        GitBaseTestCase.create_src(self)
        os.chdir(self.src_repo)
        self.make_branch(self.src_repo, 'somebranch')

    def make_branch(self, src_dir, name):
        """create a branch
        """
        subprocess.check_call(['git', 'branch', name], cwd=src_dir)

    def test_switch_local_branch(self):
        """Switch to a branch created before the clone.

        In this case, the branch already exists in local repo
        """
        target_dir = os.path.join(self.dst_dir, "to_branch")
        target_file = os.path.join(target_dir, 'tracked')
        branch = GitRepo(target_dir, self.src_repo)

        # update file from master after branching
        branch("master")
        git_write_commit(target_dir, 'tracked',
                         "last after branch", msg="last version")

        # check that no changes exists when switching from one to other
        branch('somebranch')
        self.assertFalse(branch.uncommitted_changes())
        branch('master')
        self.assertFalse(branch.uncommitted_changes())

        # modify the branch
        branch('somebranch')
        self.assertFalse(branch.uncommitted_changes())
        self.assertFalse(branch.uncommitted_changes())
        git_write_commit(target_dir, 'tracked',
                         "last after branch", msg="last version")

        git_write_commit(target_dir, 'tracked',
                         "last from branch", msg="last version")
        with open(target_file) as f:
            self.assertEqual(f.read().strip(), "last from branch")

        # switch to master
        branch('master')
        self.assertFalse(branch.uncommitted_changes())
        with open(target_file) as f:
            self.assertEqual(f.read().strip(), "last after branch")

    def test_switch_remote_branch(self):
        """Switch to a branch created after the clone.

        In this case, the branch doesn't exist in local repo
        """
        target_dir = os.path.join(self.dst_dir, "to_branch")
        branch = GitRepo(target_dir, self.src_repo)
        # update file from master after branching
        branch("master")

        # create the remote branch with some modifications
        self.make_branch(self.src_repo, 'remotebranch')
        with working_directory_keeper:
            os.chdir(self.src_repo)
            subprocess.call(['git', 'checkout', 'remotebranch'])

        git_write_commit(self.src_repo, 'tracked',
                         "last after remote branch", msg="last version")

        # switch to the remote branch and check tracked file has been updated
        branch("remotebranch")
        with open(os.path.join(target_dir, 'tracked')) as f:
            self.assertEquals(f.read().strip(), "last after remote branch")



"""VCS tests: Mercurial."""

import os
import subprocess
from ConfigParser import ConfigParser, RawConfigParser
from ..testing import COMMIT_USER_FULL
from ..testing import VcsTestCase
from ..hg import HgRepo
from ..base import UpdateError


class HgBaseTestCase(VcsTestCase):
    """Common utilities for Mercurial test cases."""

    def get_parent_node(self, repo):
        """Assert parent to be unique and return its hash code."""
        p = subprocess.Popen(['hg', '--cwd', repo,
                              'parents', '--template={node} '],
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
        nodes = p.communicate()[0].split()
        self.assertEquals(len(nodes), 1, msg="Expected only one parent")
        return nodes[0]

    def create_src(self):
        os.chdir(self.src_dir)
        subprocess.call(['hg', 'init', 'src-repo'])
        self.src_repo = os.path.join(self.src_dir, 'src-repo')
        os.chdir('src-repo')
        f = open('tracked', 'w')
        f.write("default" + os.linesep)
        f.close()
        subprocess.call(['hg', 'commit', '-A', '-m', 'initial commit',
                         '-u', COMMIT_USER_FULL])

        self.rev0 = self.get_parent_node(self.src_repo)
        subprocess.call(['hg', 'branch', 'future'])
        f = open('tracked', 'w')
        f.write("future" + os.linesep)
        f.close()
        subprocess.call(['hg', 'commit', '-m', 'in branch',
                         '-u', COMMIT_USER_FULL])
        self.rev1 = self.get_parent_node(self.src_repo)

    def make_clone(self, path, initial_rev):
        """Make a clone of the source at initial_rev.
        """
        target_dir = os.path.join(self.dst_dir, initial_rev)
        repo = HgRepo(target_dir, self.src_repo)
        repo(initial_rev)
        return repo

    def assertFutureBranch(self, repo):
        """Check that we are on the 'future' branch in target_dir repo."""
        target_dir = repo.target_dir
        self.assertTrue(os.path.isdir(target_dir))
        f = open(os.path.join(target_dir, 'tracked'))
        lines = f.readlines()
        f.close()
        self.assertEquals(lines[0].strip(), 'future')

    def assertDefaultBranch(self, repo):
        """Check that we are on the 'future' branch in target_dir repo."""
        target_dir = repo.target_dir
        self.assertTrue(os.path.isdir(target_dir))
        f = open(os.path.join(target_dir, 'tracked'))
        lines = f.readlines()
        f.close()
        self.assertEquals(lines[0].strip(), 'default')

    def assertRevision(self, branch, revno):
        p = subprocess.Popen(['hg', '--cwd', branch.target_dir,
                              'parents', '--template={rev}\n'],
                             stdout=subprocess.PIPE)
        self.assertEquals(p.communicate()[0].split(), [str(revno)])


class HgTestCase(HgBaseTestCase):

    def test_clone(self):
        repo = self.make_clone("My clone", 'default')
        self.assertDefaultBranch(repo)

    def test_clone_to_rev(self):
        """Directly clone and update to given revision."""
        repo = self.make_clone("My clone", 'future')
        self.assertFutureBranch(repo)

    def test_clean(self):
        target_dir = os.path.join(self.dst_dir, 'default')
        repo = HgRepo(target_dir, self.src_repo)

        try:
            repo.clean()
        except:
            self.fail("clean() should not fail if "
                      "clone not already done")

        repo('default')
        rc_path = os.path.join(repo.target_dir, '.hg', 'hgrc')

        # make sure that 'purge' extension is activated
        # we expect it to be installed on the system this test runs
        # note that that's the case with Debian and Red Hat families of
        # distributions
        parser = RawConfigParser()
        parser.read(rc_path)
        parser.add_section('extensions')
        parser.set('extensions', 'hgext.purge', '')
        with open(rc_path, 'w') as rc:
            parser.write(rc)

        dirty_dir = os.path.join(repo.target_dir, 'dirty')
        os.mkdir(dirty_dir)
        dirty_files = (os.path.join(repo.target_dir, 'untracked'),
                       os.path.join(dirty_dir, 'untracked2'))
        for path in dirty_files:
            with open(path, 'w') as f:
                f.write('content')
        repo.clean()
        for path in dirty_files:
            self.failIf(os.path.exists(path),
                        "Untracked file should have been removed")
        self.failIf(os.path.exists(dirty_dir),
                    "Untracked dir should have been removed")

    def test_clean_no_extension(self):
        repo = self.make_clone("clone to clean", 'default')
        rc_path = os.path.join(repo.target_dir, '.hg', 'hgrc')

        # make sure that 'purge' extension is NOT activated
        # we expect it to be installed on the system this test runs
        # note that that's the case with Debian and Red Hat families of
        # distributions
        parser = RawConfigParser()
        parser.read(rc_path)
        parser.add_section('extensions')
        parser.set('extensions', 'hgext.purge', '!')
        with open(rc_path, 'w') as rc:
            parser.write(rc)

        dirty_dir = os.path.join(repo.target_dir, 'dirty')
        os.mkdir(dirty_dir)
        dirty_files = (os.path.join(repo.target_dir, 'untracked.pyc'),
                       os.path.join(dirty_dir, 'untracked2.pyo'))
        for path in dirty_files:
            with open(path, 'w') as f:
                f.write('content')
        repo.clean()
        for path in dirty_files:
            self.failIf(os.path.exists(path),
                        "Untracked file %r should have been removed" % path)
        self.failIf(os.path.exists(dirty_dir),
                    "Untracked dir should have been removed")

    def test_update(self):
        repo = self.make_clone("clone to update", 'default')
        default_heads = repo.parents()

        repo('future')
        self.assertFutureBranch(repo)
        self.assertNotEqual(repo.parents(), default_heads)

    def test_update_same_branch(self):
        """Test that updating on a revision that we have but is a branch works.

        This should trigger a pull
        """
        repo = self.make_clone("clone to update", 'future')

        # let's make a new revision in the remote repo
        newfile = os.path.join(self.src_repo, 'newfile')
        with open(newfile, 'w') as f:
            f.write('something')
        subprocess.check_call(['hg', '--cwd', self.src_repo,
                               'commit', '-A', '-m',
                               "new commit on future branch",
                               '-u', COMMIT_USER_FULL])

        repo('future')
        self.assertFutureBranch(repo)
        self.assertRevision(repo, 2)  # would not have worked without a pull

    def test_update_fixed_rev(self):
        """Test update on a fixed rev that we already have."""
        repo = self.make_clone("clone to update", 'default')
        repo(self.rev0)
        self.assertRevision(repo, 0)

    def test_update_missing_fixed_rev(self):
        """Test update on a fixed rev that we don't have."""
        repo = self.make_clone("clone to update", 'default')
        repo(self.rev1)
        self.assertRevision(repo, 1)

    def test_have_fixed_revision(self):
        """Test some corner cases of have_fixed_revision."""
        repo = self.make_clone("clone to update", 'default')
        repo(self.rev0)
        self.assertFalse(repo.have_fixed_revision('tip'))
        self.assertFalse(repo.have_fixed_revision(''))
        self.assertTrue(repo.have_fixed_revision('0'))
        self.assertTrue(repo.have_fixed_revision(self.rev0[:12]))
        self.assertFalse(repo.have_fixed_revision(self.rev0[:2]))

    def test_hgrc_paths_update(self):
        """Method to update hgrc paths updates them and stores old values"""
        repo = self.make_clone("clone to update", 'default')
        target_dir = repo.target_dir

        # first rename
        new_src = os.path.join(self.src_dir, 'new-src-repo')
        HgRepo(target_dir, new_src).update_hgrc_paths()
        parser = ConfigParser()
        parser.read(os.path.join(target_dir, '.hg', 'hgrc'))
        self.assertEquals(parser.get('paths', 'default'), new_src)
        self.assertEquals(parser.get('paths', 'buildout_save_1'),
                          self.src_repo)

        # second rename
        new_src_2 = os.path.join(self.src_dir, 'renew-src-repo')
        HgRepo(target_dir, new_src_2).update_hgrc_paths()
        parser = ConfigParser()
        parser.read(os.path.join(target_dir, '.hg', 'hgrc'))
        self.assertEquals(parser.get('paths', 'default'), new_src_2)
        self.assertEquals(parser.get('paths', 'buildout_save_1'),
                          self.src_repo)
        self.assertEquals(parser.get('paths', 'buildout_save_2'), new_src)

    def test_url_change(self):
        """HgRepo adapts itself to changes in source URL."""
        repo = self.make_clone("clone to update", 'default')
        target_dir = repo.target_dir

        # rename and update
        new_src = os.path.join(self.src_dir, 'new-src-repo')
        os.rename(self.src_repo, new_src)
        repo = HgRepo(target_dir, new_src)
        repo('future')
        self.assertFutureBranch(repo)

    def test_uncommitted_changes(self):
        """HgRepo can detect uncommitted changes."""
        # initial cloning
        target_dir = os.path.join(self.dst_dir, "clone to update")
        repo = HgRepo(target_dir, self.src_repo)
        repo('default')

        self.assertFalse(repo.uncommitted_changes())

        f = open(os.path.join(target_dir, 'tracked'), 'w')
        f.write('mod')
        f.close()

        self.assertTrue(repo.uncommitted_changes())

    def test_failed(self):
        target_dir = os.path.join(self.dst_dir, "My clone")
        repo = HgRepo(target_dir, '/does-not-exit')
        self.assertRaises(subprocess.CalledProcessError,
                          repo.get_update, 'default')


class HgOfflineTestCase(HgBaseTestCase):

    def make_clone(self, path, initial_rev):
        """Make a local branch of the source at initial_rev and forbid pulls.
        """
        repo = HgBaseTestCase.make_clone(self, path, initial_rev)
        repo.offline = True

        def _pull():
            raise UpdateError("Should not pull !")
        repo._pull = _pull
        return repo

    def test_update_fixed_rev_by_node(self):
        """[offline mode] test update on a fixed rev that we already have."""
        repo = self.make_clone("clone to update", 'default')
        repo(self.rev0)
        self.assertRevision(repo, 0)

    def test_update_fixed_rev_by_tag(self):
        """[offline mode] test update on a tag that we already have."""
        repo = self.make_clone("clone to update", 'default')
        subprocess.check_call(['hg', '--cwd', repo.target_dir,
                               'tag', '-r', self.rev0, 'first'])
        repo('first')
        self.assertRevision(repo, 0)

    def test_update_missing_fixed_rev(self):
        """[offline mode] update on a rev that we don't have raises."""
        repo = self.make_clone("clone to update", 'default')
        self.assertRaises(UpdateError, repo, self.rev1)
        # must also fail as branch
        self.assertRaises(UpdateError, repo, 'future')

    def test_update_missing_tag(self):
        """[offline mode] update on a tag that we don't have raises."""
        repo = self.make_clone("clone to update", 'default')
        subprocess.check_call(['hg', '--cwd', self.src_repo,
                               'tag', '-r', self.rev0, 'in-src-only'])
        repo(self.rev0)
        self.assertRaises(UpdateError, repo, 'in-src-only')

    def test_update_branch_head(self):
        """[offline mode] update on branch head should not pull"""
        repo = self.make_clone("clone to update", '0')
        repo('default')
        self.assertRevision(repo, '0')



import os
import sys
import subprocess
import shutil
import tempfile
from ConfigParser import ConfigParser, NoOptionError
from anybox.recipe.openerp.server import BaseRecipe
from anybox.recipe.openerp.base import main_software
from anybox.recipe.openerp.base import GP_VCS_EXTEND_DEVELOP
from anybox.recipe.openerp.testing import RecipeTestCase
from ..testing import COMMIT_USER_FULL

TEST_DIR = os.path.dirname(__file__)


class TestingRecipe(BaseRecipe):
    """A subclass with just enough few defaults for unit testing."""

    archive_filenames = {'6.1': 'blob-%s.tgz'}
    archive_nightly_filenames = {'6.1': '6-1-nightly-%s.tbz'}


class TestBaseRecipe(RecipeTestCase):

    def make_recipe(self, name='openerp', **options):
        self.recipe = TestingRecipe(self.buildout, name, options)

    def get_source_type(self):
        return self.recipe.sources[main_software][0]

    def get_source_url(self):
        return self.recipe.sources[main_software][1][0]

    def assertDownloadUrl(self, url):
        """Assert that main software is 'downloadable' with given url."""
        source = self.recipe.sources[main_software]
        self.assertEquals(source[0], 'downloadable')
        self.assertEquals(source[1][0], url)

    def test_version_release_6_1(self):
        self.make_recipe(version='6.1-1')

        recipe = self.recipe
        self.assertEquals(recipe.version_wanted, '6.1-1')
        self.assertDownloadUrl(
            'http://nightly.openerp.com/6.1/releases/blob-6.1-1.tgz')

    def test_version_nightly_6_1(self):
        self.make_recipe(version='nightly 6.1 1234-5')

        self.assertDownloadUrl(
            'http://nightly.openerp.com/6.1/nightly/src/'
            '6-1-nightly-1234-5.tbz')

    def test_version_bzr_6_1(self):
        self.make_recipe(
            version='bzr lp:openobject-server/6.1 openerp-6.1 last:1')

        recipe = self.recipe
        self.assertEquals(self.get_source_type(), 'bzr')
        self.assertEquals(self.get_source_url(), 'lp:openobject-server/6.1')
        self.assertEquals(recipe.openerp_dir,
                          os.path.join(recipe.parts, 'openerp-6.1'))

    def test_version_local(self):
        local_path = 'path/to/local/version'
        self.make_recipe(version='local ' + local_path)
        recipe = self.recipe
        self.assertEquals(self.get_source_type(), 'local')
        self.assertTrue(recipe.openerp_dir.endswith(local_path))

    def test_version_url(self):
        url = 'http://download.example/future/openerp-12.0.tgz'
        self.make_recipe(version='url ' + url)
        recipe = self.recipe
        self.assertDownloadUrl(url)
        self.assertEquals(recipe.archive_filename, 'openerp-12.0.tgz')

    def test_base_url(self):
        self.make_recipe(version='6.1-1',
                         base_url='http://example.org/openerp')
        self.assertDownloadUrl('http://example.org/openerp/blob-6.1-1.tgz')

    def test_base_url_nightly(self):
        self.make_recipe(version='nightly 6.1 1234-5',
                         base_url='http://example.org/openerp')
        self.assertDownloadUrl(
            'http://example.org/openerp/6-1-nightly-1234-5.tbz')

    def test_buildout_cfg_name(self):
        self.make_recipe(version='6.1-1')
        bcn = self.recipe.buildout_cfg_name
        self.assertEquals(bcn(), 'buildout.cfg')
        self.assertEquals(bcn(('-D', 'install', 'openerp')), 'buildout.cfg')
        self.assertEquals(bcn(('-c', '6.1.cfg')), '6.1.cfg')
        self.assertEquals(bcn(('--config', '6.1.cfg')), '6.1.cfg')
        self.assertEquals(bcn(('-o', '--config', '6.1.cfg')), '6.1.cfg')
        self.assertEquals(bcn(('--config=6.1.cfg',)), '6.1.cfg')
        self.assertEquals(bcn(('--config=6.1.cfg', '-o')), '6.1.cfg')

    def test_parse_addons_revisions(self):
        """Test both parse_addons and parse_revisions."""
        self.make_recipe(version='bzr lp:openobject-server server last:1'
                         ' option=option')
        recipe = self.recipe

        recipe.parse_revisions(dict(revisions='1234'))
        self.assertEquals(recipe.sources[main_software],
                          ('bzr', ('lp:openobject-server', '1234'),
                           dict(option='option')))

        recipe.parse_addons(
            dict(addons='hg http://some/repo addons-specific default opt=spam')
        )
        self.assertEquals(recipe.sources['addons-specific'],
                          ('hg', ('http://some/repo', 'default'),
                           {'opt': 'spam'}))

        recipe.parse_revisions(dict(revisions='1111\naddons-specific 1.0'))
        self.assertEquals(recipe.sources['addons-specific'],
                          ('hg', ('http://some/repo', '1.0'), {'opt': 'spam'}))
        self.assertEquals(recipe.sources[main_software][1][1], '1111')

        # with a comment
        recipe.parse_revisions(dict(revisions='1112 ; main software'
                                    '\naddons-specific 1.0'))
        self.assertEquals(recipe.sources[main_software][1][1], '1112')

    def build_babel_egg(self):
        """build an egg for fake babel in buildout's eggs directory."""
        subprocess.check_call(
            [sys.executable,
             os.path.join(TEST_DIR, 'fake_babel', 'setup.py'),
             'bdist_egg',
             '-d', self.recipe.b_options['eggs-directory'],
             '-b', os.path.join(self.buildout_dir, 'build')])

    def develop_babel(self):
        """Develop fake babel in buildout's directory"""
        subprocess.check_call(
            [sys.executable,
             os.path.join(TEST_DIR, 'fake_babel', 'setup.py'),
             'develop',
             '-d', self.recipe.b_options['develop-eggs-directory'],
             '-b', os.path.join(self.buildout_dir, 'build')])

    def test_freeze_egg_versions(self):
        """Test that an egg requirement is properly dumped with its version.
        """
        conf = ConfigParser()
        conf.add_section('freeze')
        self.make_recipe(version='6.1-1')
        self.build_babel_egg()
        self.recipe.options['eggs'] = 'Babel'
        self.recipe.install_requirements()  # to get 'ws' attribute
        self.recipe._freeze_egg_versions(conf, 'freeze')
        try:
            version = conf.get('freeze', 'Babel')
        except NoOptionError:
            self.fail("Expected version of Babel egg not dumped !")
        self.assertEquals(version, '0.123')

    def test_freeze_egg_versions_merge(self):
        """Test that freezing of egg versions keeps eggs already dumped.

        very much similar to test_freeze_egg_versions.
        """
        conf = ConfigParser()
        conf.add_section('freeze')
        conf.set('freeze', 'some.distribution', '1.0alpha')
        self.make_recipe(version='6.1-1')
        self.build_babel_egg()
        self.recipe.options['eggs'] = 'Babel'
        self.recipe.install_requirements()  # to get 'ws' attribute
        self.recipe._freeze_egg_versions(conf, 'freeze')
        try:
            version = conf.get('freeze', 'some.distribution')
        except NoOptionError:
            self.fail("Expected version of 'some.distribution' not kept !")
        self.assertEquals(version, '1.0alpha')

    def test_freeze_egg_versions_develop(self):
        """Test that a developped requirement is not dumped in [versions].
        """
        conf = ConfigParser()
        conf.add_section('freeze')
        self.make_recipe(version='6.1-1')
        self.recipe.develop(os.path.join(TEST_DIR, 'fake_babel'))
        self.recipe.options['eggs'] = 'Babel'
        self.recipe.install_requirements()  # to get 'ws' attribute
        self.recipe._freeze_egg_versions(conf, 'freeze')
        self.assertRaises(NoOptionError, conf.get, 'freeze', 'Babel')

    def test_freeze_vcs_source(self):
        self.make_recipe(version='6.1-1')
        b_dir = self.recipe.buildout_dir
        repo_path = os.path.join(b_dir, 'custom')
        subprocess.check_call(['hg', 'init', repo_path])
        with open(os.path.join(repo_path, 'somefile'), 'w') as f:
            f.write('content')
        subprocess.check_call(['hg', '--cwd', repo_path,
                               'commit', '-A', '-m', 'somerev',
                               '-u', COMMIT_USER_FULL,
                               ])

        rev = self.recipe._freeze_vcs_source('hg', repo_path)
        hg = subprocess.Popen(['hg', '--cwd', repo_path, 'diff', '-r', rev],
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE)
        out, err = hg.communicate()
        if hg.returncode or err:
            self.fail("Invalid extracted revision: %r" % rev)
        self.assertEquals(out, '', 'Extracted revision shows some diff')

    def test_clean(self):
        """Test clean for local server & addons and base class vcs addons.
        """
        self.make_recipe(
            version='local server-dir',
            addons=os.linesep.join((
                'local addon-dir',
                'fakevcs http://some/repo vcs-addons revspec')),
            clean='true')

        b_dir = self.recipe.buildout_dir
        for d in (['server-dir'], ['server-dir', 'a'],
                  ['addon-dir'],
                  ['addon-dir', 'a'], ['addon-dir', 'b'],
                  ['vcs-addons'],
                  ['vcs-addons', 'a']):
            os.mkdir(os.path.join(b_dir, *d))
        for path in (['server-dir', 'a', 'x.pyc'],
                     ['addon-dir', 'a', 'x.pyc'],
                     ['addon-dir', 'b', 'x.pyo'],
                     ['addon-dir', 'b', 'other-file'],
                     ['vcs-addons', 'a', 'x.pyc'],
                     ['vcs-addons', 'a', 'x.py'],
                     ):
            with open(os.path.join(b_dir, *path), 'w') as f:
                f.write("content")
        self.recipe.retrieve_main_software()
        self.recipe.retrieve_addons()

        for path, expected in (
                (('server-dir',), True),
                (('server-dir', 'a', 'x.pyc'), False),
                (('server-dir', 'a'), False),
                (('vcs-addons', 'a', 'x.py'), True),
                (('vcs-addons', 'a', 'x.pyc'), False),
                (('addon-dir',), True),
                (('addon-dir', 'a', 'x.pyc'), False),
                (('addon-dir', 'a'), False),
                (('addon-dir', 'b'), True),
                (('addon-dir', 'b', 'x.pyo'), False),
                (('addon-dir', 'b', 'other-file'), True)):
            self.assertEquals(os.path.exists(os.path.join(b_dir, *path)),
                              expected)

    def test_clean_vcs_server(self):
        """Test clean for base class vcs server."""
        self.make_recipe(
            version='fakevcs http://some/where server-dir somerev',
            clean='true')

        b_dir = self.recipe.buildout_dir
        server_path = os.path.join(b_dir, 'parts', 'server-dir')
        os.makedirs(os.path.join(server_path, 'a'))
        for path in (['a', 'x.pyc'],
                     ['a', 'x.py'],
                     ):
            with open(os.path.join(server_path, *path), 'w') as f:
                f.write("content")
        self.recipe.retrieve_main_software()
        self.recipe.retrieve_addons()

        for path, expected in (
                (('a',), True),
                (('a', 'x.pyc'), False),
                (('a', 'x.py'), True)):
            self.assertEquals(os.path.exists(os.path.join(server_path, *path)),
                              expected)

    def test_freeze_vcs_source_dirty(self):
        self.make_recipe(version='6.1-1')
        b_dir = self.recipe.buildout_dir
        repo_path = os.path.join(b_dir, 'custom')
        subprocess.check_call(['hg', 'init', repo_path])
        with open(os.path.join(repo_path, 'somefile'), 'w') as f:
            f.write('content')
        subprocess.check_call(['hg', '--cwd', repo_path,
                               'commit', '-A', '-m', 'somerev',
                               '-u', COMMIT_USER_FULL,
                               ])

        self.recipe.local_modifications = []
        # modification on tracked file
        with open(os.path.join(repo_path, 'somefile'), 'w') as f:
            f.write('changed content')
        self.recipe._freeze_vcs_source('hg', repo_path)
        self.assertTrue(bool(self.recipe.local_modifications))

        subprocess.check_call(['hg', '--cwd', repo_path, 'revert', '--all'])

        # untracked file
        self.recipe.local_modifications = []
        with open(os.path.join(repo_path, 'untracked'), 'w') as f:
            f.write('something else')
        self.recipe._freeze_vcs_source('hg', repo_path)
        self.assertTrue(bool(self.recipe.local_modifications))

    def test_prepare_frozen_buildout(self):
        self.make_recipe(version='6.1-1')
        conf = ConfigParser()
        self.recipe._prepare_frozen_buildout(conf)
        self.assertTrue('buildout' in conf.sections())

    def test_prepare_frozen_buildout_gp_vcsdevelop(self):
        self.make_recipe(version='6.1-1')
        self.recipe.b_options[GP_VCS_EXTEND_DEVELOP] = (
            "fakevcs+http://example.com/aeroolib#egg=aeroolib")

        conf = ConfigParser()
        self.recipe._prepare_frozen_buildout(conf)
        extends_develop = conf.get('buildout', GP_VCS_EXTEND_DEVELOP)
        self.assertEquals(extends_develop.strip(),
                          "fakevcs+http://example.com/aeroolib@fakerev"
                          "#egg=aeroolib")

    def test_prepare_frozen_buildout_gp_vcsdevelop_already_fixed(self):
        """Test that prepare_frozen_buildout understands existing pinning.

        One might say that we souldn't touch an existing revision pinning, but
        a difference can arise from a tag resolution, or simply someone
        manually updating the repo. In all cases, the instrospected revision
        will be used.
        """
        self.make_recipe(version='6.1-1')
        self.recipe.b_options[GP_VCS_EXTEND_DEVELOP] = (
            "fakevcs+http://example.com/aeroolib@somerev#egg=aeroolib")

        conf = ConfigParser()
        self.recipe._prepare_frozen_buildout(conf)
        extends_develop = conf.get('buildout', GP_VCS_EXTEND_DEVELOP)
        self.assertEquals(extends_develop.strip(),
                          "fakevcs+http://example.com/aeroolib@fakerev"
                          "#egg=aeroolib")


class TestExtraction(RecipeTestCase):

    def setUp(self):
        super(TestExtraction, self).setUp()
        self.extract_target_dir = tempfile.mkdtemp('test_recipe_extract')

    def tearDown(self):
        shutil.rmtree(self.extract_target_dir)
        super(TestExtraction, self).tearDown()

    def make_recipe(self, name='openerp', **options):
        self.recipe = TestingRecipe(self.buildout, name, options)

    def test_prepare_extracted_buildout(self):
        self.make_recipe(version='6.1-1')
        conf = ConfigParser()
        self.recipe._prepare_extracted_buildout(conf, self.extract_target_dir)
        self.assertTrue('buildout' in conf.sections())

    def test_extract_addons(self):
        """Test extract_downloads_to about addons ('local' server version).
        """
        target_dir = self.extract_target_dir
        addons = ['local specific',
                  'fakevcs http://some/repo vcs-addons revspec']
        self.make_recipe(version='local mainsoftware',
                         addons=os.linesep.join(addons))

        conf = ConfigParser()
        extracted = set()
        self.recipe._extract_sources(conf, target_dir, extracted)
        addons_opt = set(conf.get('openerp', 'addons').split(os.linesep))
        self.assertEquals(addons_opt,
                          set(('local vcs-addons', 'local specific')))
        self.assertEquals(extracted,
                          set([os.path.join(target_dir, 'vcs-addons')]))

        # no need to override revisions
        self.assertRaises(NoOptionError, conf.get, 'openerp', 'revisions')

        # testing that archival took place for fakevcs, but not for local

        self.failIf(os.path.exists(os.path.join(target_dir, 'specific')),
                    "Local addons dir should not have been extracted")
        # get_update having not been called, it is expected to have the
        # default revision 'fakerev', instead of 'revspec'.
        with open(os.path.join(target_dir, 'vcs-addons',
                               '.fake_archival.txt')) as f:
            self.assertEquals(f.read(), 'fakerev')

    def test_extract_addons_revisions(self):
        """Test extract_downloads_to about revisions overriding.

        In case the source buildout uses the revisions option, it must be
        overridden in the extracted one because it does not make sense with
        the 'local' scheme.
        """
        target_dir = self.extract_target_dir
        addons = ['local specific',
                  'fakevcs http://some/repo vcs-addons revspec']
        self.make_recipe(version='local mainsoftware',
                         addons=os.linesep.join(addons),
                         revisions='vcs-addons 213')

        conf = ConfigParser()
        extracted = set()
        self.recipe._extract_sources(conf, target_dir, extracted)
        self.assertEquals(conf.get('openerp', 'revisions').strip(), '')

    def test_prepare_extracted_buildout_gp_vcsdevelop(self):
        self.make_recipe(version='6.1-1')
        self.recipe.b_options[GP_VCS_EXTEND_DEVELOP] = (
            "fakevcs+http://example.com/aeroolib#egg=aeroolib")
        self.recipe.b_options['develop'] = os.path.join(
            self.recipe.buildout_dir, 'simple_develop')

        conf = ConfigParser()
        self.recipe._prepare_extracted_buildout(conf, self.extract_target_dir)
        extends_develop = conf.get('buildout', GP_VCS_EXTEND_DEVELOP)
        self.assertEquals(extends_develop.strip(), '')
        develop = conf.get('buildout', 'develop').split(os.linesep)
        self.assertEquals(set(d for d in develop if d),
                          set(['aeroolib', 'simple_develop']))

        # extraction has been done
        target = os.path.join(self.extract_target_dir, 'aeroolib')
        self.assertTrue(os.path.exists(target))
        with open(os.path.join(target, '.fake_archival.txt')) as f:
            self.assertEquals(f.read(), 'fakerev')



"""Test VCS methods (common part).

These tests depend on system executables, such as 'hg', 'bzr', etc.
"""

import os
import subprocess

from .. import get_update
from .. import testing
from .. import SUPPORTED
from ..base import UpdateError
from ..hg import HgRepo


class CommonTestCase(testing.VcsTestCase):
    """Test methods that are common among the different repository classes."""

    def create_src(self):
        """We use HgRepo to test the common features."""

        os.chdir(self.src_dir)
        subprocess.call(['hg', 'init', 'src-repo'])
        self.src_repo = os.path.join(self.src_dir, 'src-repo')
        os.chdir('src-repo')
        f = open('tracked', 'w')
        f.write("default" + os.linesep)
        f.close()
        subprocess.call(['hg', 'add'])
        subprocess.call(['hg', 'commit', '-m', 'initial commit',
                         '-u', testing.COMMIT_USER_FULL])

    def test_unknown(self):
        self.assertRaises(ValueError,
                          get_update, 'unknown', '', '', 'default')

    def test_retry(self):
        """With a repo class that fails updates, retry works."""

        class HgRepoFailsUpdates(HgRepo):
            def get_update(self, revision):
                if os.path.exists(self.target_dir):
                    raise UpdateError(1, ['error'])
                HgRepo.get_update(self, revision)

        SUPPORTED['hg_fails_updates'] = HgRepoFailsUpdates
        repo_path = os.path.join(self.dst_dir, "clone")

        get_update('hg_fails_updates', repo_path, self.src_repo, 'default')
        self.assertTrue(os.path.isdir(repo_path))

        # without the retry option
        self.assertRaises(UpdateError, get_update, 'hg_fails_updates',
                          repo_path, self.src_repo, 'default')

        # now the retry
        get_update('hg_fails_updates', repo_path, self.src_repo, 'default',
                   clear_retry=True)

        # no such wild retry in offline mode
        self.assertRaises(UpdateError, get_update, 'hg_fails_updates',
                          repo_path, self.src_repo, 'default', offline=True)



import os
import sys
import re
import subprocess
from contextlib import contextmanager
import logging
logger = logging.getLogger(__name__)


MAJOR_VERSION_RE = re.compile(r'(\d+)[.](\d*)(\w*)')


class WorkingDirectoryKeeper(object):
    """A context manager to get back the working directory as it was before.

    If you want to stack working directory keepers, you need a new instance
    for each stage.
    """

    active = False

    def __enter__(self):
        if self.active:
            raise RuntimeError("Already in a working directory keeper !")
        self.wd = os.getcwd()
        self.active = True

    def __exit__(self, *exc_args):
        os.chdir(self.wd)
        self.active = False

working_directory_keeper = WorkingDirectoryKeeper()


@contextmanager
def use_or_open(provided, path, *open_args):
    """A context manager to use an open file if not None or open one.

    Useful for code that should be unit-testable, but work on a default file if
    None is passed.
    """
    if provided is not None:
        yield provided
    else:
        with open(path, *open_args) as f:
            yield f


def major_version(version_string):
    """The least common denominator of OpenERP versions : two numbers.

    OpenERP version numbers are a bit hard to compare if we consider nightly
    releases, bzr versions etc. It's almost impossible to compare them without
    an a priori knowledge of release dates and revisions.

    Here are some examples::

       >>> major_version('1.2.3-foo.bar')
       (1, 2)
       >>> major_version('6.1-20121003-233130')
       (6, 1)
       >>> major_version('7.0alpha')
       (7, 0)

    Beware, the packaging script does funny things, such as labeling current
    nightlies as 6.2-date-time whereas version_info is (7, 0, 0, ALPHA)
    We can in recipe code check for >= (6, 2), that's not a big issue.

    Regarding OpenERP saas releases (e.g. 7.saas~1) that are short-lived stable
    versions between two "X.0" LTS releases, the second version number does not
    contain a numeric value. The value X.5 will be returned (e.g. 7.5)::

       >>> major_version('7.saas~1')
       (7, 5)

    """

    m = MAJOR_VERSION_RE.match(version_string)

    if m is None:
        raise ValueError("Unparseable version string: %r" % version_string)

    major = int(m.group(1))
    minor = m.group(2)
    if not minor and m.group(3).startswith('saas'):
        return major, 5

    try:
        return major, int(minor)
    except TypeError:
        raise ValueError(
            "Unrecognized second version segment in %r" % version_string)


def mkdirp(path):
    """Same as mkdir -p."""
    if not os.path.exists(path):
        parent, name = os.path.split(path)
        mkdirp(parent)
        os.mkdir(path)


def is_object_file(filename):
    """True if given filename is a python object file."""
    return filename.endswith('.pyc') or filename.endswith('.pyo')


def clean_object_files(directory):
    """Recursively remove object files in given directory.

    Also remove resulting empty directories.
    """
    dirs_to_remove = []
    for dirpath, dirnames, filenames in os.walk(directory, topdown=False):
        to_delete = [os.path.join(dirpath, f)
                     for f in filenames if is_object_file(f)]
        if not dirnames and len(to_delete) == len(filenames):
            dirs_to_remove.append(dirpath)
        for p in to_delete:
            try:
                os.unlink(p)
            except:
                logger.exception("Error attempting to unlink %r. "
                                 "Proceeding anyway.", p)
    for d in dirs_to_remove:
        try:
            os.rmdir(d)
        except:
            logger.exception("Error attempting to rmdir %r",
                             "Proceeding anyway.", p)


def check_output(*popenargs, **kwargs):
    r"""Backport of subprocess.check_output from python 2.7.

    Example (this doctest would be more readable with ELLIPSIS, but
    that's good enough for today):

    >>> out = check_output(["ls", "-l", "/dev/null"])
    >>> out.startswith('crw-rw-rw')
    True

    The stdout argument is not allowed as it is used internally.
    To capture standard error in the result, use stderr=STDOUT.

    >>> os.environ['LANG'] = 'C'  # for uniformity of error msg
    >>> err = check_output(["/bin/sh", "-c",
    ...               "ls -l non_existent_file ; exit 0"],
    ...              stderr=subprocess.STDOUT)
    >>> err.strip().endswith("No such file or directory")
    True
    """

    if sys.version >= (2, 7):
        return subprocess.check_output(*popenargs, **kwargs)
    if 'stdout' in kwargs:
        raise ValueError('stdout argument not allowed, it will be overridden.')

    process = subprocess.Popen(stdout=subprocess.PIPE, *popenargs, **kwargs)
    output, unused_err = process.communicate()
    retcode = process.poll()
    if retcode:
        cmd = kwargs.get("args")
        if cmd is None:
            cmd = popenargs[0]
        # in python 2.6, CalledProcessError.__init__ does not have output kwarg
        exc = subprocess.CalledProcessError(retcode, cmd)
        exc.output = output
        raise exc
    return output



# -*- coding: utf-8 -*-

##############################################################################
#
#    OpenERP, Open Source Management Solution
#    Copyright (C) 2004-2009 Tiny SPRL (<http://tiny.be>).
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################
#
# anybox.recipe.openerp is also under AGPL v3+
#

"""
Copied from OpenERP 6.1's openerp-server startup script.

This makes utilitary functions available to the session module.
"""

import logging
import os
import sys

try:
    import openerp
except:
    # this file must stay importable by nose tests
    pass
else:
    __author__ = openerp.release.author
    __version__ = openerp.release.version

    if openerp.release.version_info < (6, 1):
        raise RuntimeError("Interpreter and scripts not compatible "
                           "with OpenERP < 6.1")

# Also use the `openerp` logger for the main script.
_logger = logging.getLogger('openerp')


def check_root_user():
    """ Exit if the process's user is 'root' (on POSIX system)."""
    if os.name == 'posix':
        import pwd
        if pwd.getpwuid(os.getuid())[0] == 'root':
            sys.stderr.write("Running as user 'root' is a security risk, "
                             "aborting.\n")
            sys.exit(1)


def check_postgres_user():
    """ Exit if the configured database user is 'postgres'.

    This function assumes the configuration has been initialized.
    """
    config = openerp.tools.config
    if config['db_user'] == 'postgres':
        sys.stderr.write("Using the database user 'postgres' is a "
                         "security risk, aborting.")
        sys.exit(1)


def report_configuration():
    """ Log the server version and some configuration values.

    This function assumes the configuration has been initialized.
    """
    config = openerp.tools.config
    _logger.info("OpenERP version %s", __version__)
    for name, value in [('addons paths', config['addons_path']),
                        ('database hostname',
                         config['db_host'] or 'localhost'),
                        ('database port', config['db_port'] or '5432'),
                        ('database user', config['db_user'])]:
        _logger.info("%s: %s", name, value)



import os
import logging
import subprocess
import urlparse
import urllib
from StringIO import StringIO

from ..utils import use_or_open
from ..utils import working_directory_keeper
from ..utils import check_output
from .base import SUBPROCESS_ENV
from .base import BaseRepo
from .base import update_check_call
from .base import clone_check_call
from .base import UpdateError
from .base import CloneError

logger = logging.getLogger(__name__)

try:
    from bzrlib.plugins.launchpad.lp_directory import LaunchpadDirectory
except ImportError:
    LPDIR = None
else:
    LPDIR = LaunchpadDirectory()


class BzrBranch(BaseRepo):
    """Represent a Bazaar branch tied to a reference branch."""

    vcs_control_dir = '.bzr'

    def __init__(self, *a, **kw):
        super(BzrBranch, self).__init__(*a, **kw)
        if self.options.get('bzr-init') == "ligthweight-checkout":
            logger.warn("The 'ligthweight-checkout' *misspelling* is "
                        "deprecated as of version 1.7.1 of this buildout "
                        "recipe. "
                        "Please fix it as 'lightweight-checkout', as it will "
                        "probably disappear in version 1.8.")
            self.options['bzr-init'] = 'lightweight-checkout'

        if self.url.startswith('lp:'):
            if LPDIR is None:
                raise RuntimeError(
                    "To use launchpad locations (lp:), bzrlib must be "
                    "importable. Please also take care that it's the same "
                    "or working exactly as the one behind the bzr executable")

            # first arg (name) of look_up is acturally ignored
            url = LPDIR.look_up('', self.url)
            parsed = list(urlparse.urlparse(url))
            parsed[2] = urllib.quote(parsed[2])
            self.url = urlparse.urlunparse(parsed)

    def conf_file_path(self):
        return os.path.join(self.target_dir, '.bzr', 'branch', 'branch.conf')

    def parse_conf(self, from_file=None):
        """Return a dict of paths from standard conf (or the given file-like)

        Reference: http://doc.bazaar.canonical.com/bzr.0.18/configuration.htm

        >>> from pprint import pprint
        >>> branch = BzrBranch('', '')
        >>> pprint(branch.parse_conf(StringIO(os.linesep.join([
        ...        "parent_location = /some/path",
        ...        "submit_location = /other/path"]))))
        {'parent_location': '/some/path', 'submit_location': '/other/path'}
        """
        with use_or_open(from_file, self.conf_file_path()) as conffile:
            return dict((name.strip(), url.strip())
                        for name, url in (
                            line.split('=', 1) for line in conffile
                            if not line.startswith('#') and '=' in line))

    def write_conf(self, conf, to_file=None):
        """Write counterpart to read_conf (see docstring of read_conf)
        """
        lines = ('%s = %s' % (k, v) + os.linesep
                 for k, v in conf.items())
        with use_or_open(to_file, self.conf_file_path(), 'w') as conffile:
            conffile.writelines(lines)

    def update_conf(self):
        try:
            conf = self.parse_conf()
        except IOError:
            logger.error("Cannot read branch.conf of Bazaar branch at %s "
                         "Proceeding anyway, remote URLs shall not be "
                         "updated if needed. Please check that the branch "
                         "is in good shape.", self.target_dir)
            return

        old_parent = conf['parent_location']
        if old_parent == self.url:
            return
        count = 1
        while True:
            save = 'buildout_save_parent_location_%d' % count
            if save not in conf:
                conf[save] = old_parent
                break
            count += 1
        conf['parent_location'] = self.url
        self.write_conf(conf)

    def uncommitted_changes(self):
        """True if we have uncommitted changes."""
        return bool(check_output(['bzr', 'status', self.target_dir],
                                 env=SUBPROCESS_ENV))

    def parents(self):
        """Return current revision.

        Must be used in conjunction with uncommitted_changes to be sure
        that we are indeed on this revision
        """

        return [check_output(['bzr', 'revno', '--tree', self.target_dir],
                             env=SUBPROCESS_ENV).strip()]

    def clean(self):
        if not os.path.exists(self.target_dir):
            # not branched yet, there's nothing to clean
            return
        with working_directory_keeper:
            os.chdir(self.target_dir)
            subprocess.check_call(['bzr', 'clean-tree',
                                   '--ignored', '--force'])

    def _update(self, revision):
        """Update existing branch at target dir to given revision.

        raise UpdateError in case of problems."""

        update_check_call(['bzr', 'up', '-r', revision, self.target_dir],
                          env=SUBPROCESS_ENV)
        logger.info("Updated %r to revision %s", self.target_dir, revision)

    def get_revid(self, revision):
        with working_directory_keeper:
            os.chdir(self.target_dir)
            try:
                log = check_output(
                    ['bzr', 'log', '--show-ids', '-r', revision],
                    env=SUBPROCESS_ENV)
            except subprocess.CalledProcessError as exc:
                if exc.returncode != 3:
                    raise
                raise LookupError(
                    "could not find revision id for %r" % revision)

            prefix = 'revision-id:'
            for line in log.split(os.linesep):
                if line.startswith(prefix):
                    return line[len(prefix):].strip()
            raise LookupError("could not find revision id for %r" % revision)

    def is_fixed_revision(self, revstr):
        """True iff the given revision string is for a fixed revision."""

        revstr = revstr.strip()  # one never knows
        if not revstr or revstr.startswith('last:'):
            return False
        try:
            revno = int(revstr)
        except ValueError:
            return True  # a string is either a tag or a revid: spec
        else:
            return revno >= 0

    def get_update(self, revision):
        """Ensure that target_dir is a branch of url at specified revision.

        If target_dir already exists, does a simple pull.
        Offline-mode: no branch nor pull, but update.
        In all cases, an attempt to update is performed before any pull
        """
        target_dir = self.target_dir
        offline = self.offline
        clear_locks = self.clear_locks

        if not os.path.exists(target_dir):
            try:
                self._branch(revision)
            except CloneError:
                if not revision:
                    raise
                logger.warning("First attempt of branching to %s at revision "
                               "%r failed. Retrying in two steps.", target_dir,
                               revision)
                # it really happens, see
                # https://bugs.launchpad.net/anybox.recipe.openerp/+bug/1204573
                self._branch(None)
                self._update(revision)

        else:
            # TODO what if bzr source is actually local fs ?
            if clear_locks:
                yes = StringIO()
                yes.write('y')
                yes.seek(0)
                logger.info("Break-lock for branch %s ...", target_dir)
                # GR newer versions of bzr have a --force option, but this call
                # works also for older ones (fortunately we don't need a pty)
                p = subprocess.Popen(['bzr', 'break-lock', target_dir],
                                     subprocess.PIPE)
                out, err = p.communicate(input='y')
                if p.returncode != 0:
                    raise subprocess.CalledProcessError(
                        p.returncode, repr(['bzr', 'break-lock', target_dir]))

            self.update_conf()

            init_opt = self.options.get('bzr-init')
            if self.is_fixed_revision(revision):
                if (offline and init_opt == 'lightweight-checkout'):
                    logger.warning("Offline mode, no update for lightweight "
                                   "checkout at %s on revision %r",
                                   self.target_dir, revision)
                    return
                try:
                    self._update(revision)
                    return
                except UpdateError:
                    if offline:
                        raise

            if offline:
                logger.info("Offline mode, no pull for revision %r", revision)
            else:
                self._pull()

            if not (offline and init_opt in ('stacked-branch',
                                             'lightweight-checkout')):
                self._update(revision)

    def _branch(self, revision):
        """ Branch or checkout remote repository
        """
        target_dir = self.target_dir
        url = self.url
        offline = self.offline
         # TODO case of local url ?
        if offline:
            raise IOError(
                "bzr branch %s does not exist; cannot branch it from "
                "%s (offline mode)" % (target_dir, url))

        options = self.options
        if "bzr-init" in options and "bzr-stacked-branches" in options:
            raise Exception(
                "Both options 'bzr-init' and "
                "'bzr-stacked-branches' are mutually exclusive. "
                "Prefer 'bzr-init'.")
        default = "branch"

        if "bzr-stacked-branches" in options:
            logger.warning("'bzr-stacked-branches' option is deprecated. "
                           "Replace by bzr-init=stacked-branch")
            default = "stacked-branch"

        bzr_opt = options.get("bzr-init", default)
        branch_cmd = ['bzr']
        if bzr_opt == "branch":
            branch_cmd.append("branch")
            logger.info("Branching %s ...", url)
        elif bzr_opt == "stacked-branch":
            branch_cmd.extend(["branch", "--stacked"])
            logger.info("Stacked branching %s ...", url)
        elif bzr_opt == "lightweight-checkout":
            branch_cmd.extend(["checkout", "--lightweight"])
            logger.info("Lightweight checkout %s ...", url)
        else:
            raise Exception("Unsupported option %r" % bzr_opt)

        if revision:
            branch_cmd.extend(['-r', revision])

        branch_cmd.extend([url, target_dir])
        clone_check_call(branch_cmd, env=SUBPROCESS_ENV)

    def _pull(self):
        if self.options.get('bzr-init') == 'lightweight-checkout':
            logger.info("Update lightweight checkout at %s ...",
                        self.target_dir)
            update_check_call(['bzr', 'update', self.target_dir],
                              env=SUBPROCESS_ENV)
        else:
            logger.info("Pull for branch %s ...", self.target_dir)
            update_check_call(['bzr', 'pull', '-d', self.target_dir],
                              env=SUBPROCESS_ENV)

    def archive(self, target_path):
        with working_directory_keeper:
            os.chdir(self.target_dir)
            subprocess.check_call(['bzr', 'export', target_path])



"""Utilities for unit tests."""
import os
import unittest
import sys
import shutil
from tempfile import mkdtemp
from zc.buildout.easy_install import Installer
from . import vcs
from . import utils

COMMIT_USER_NAME = 'Test'
COMMIT_USER_EMAIL = 'test@example.org'
COMMIT_USER_FULL = '%s %s' % (COMMIT_USER_NAME, COMMIT_USER_EMAIL)


class FakeRepo(vcs.base.BaseRepo):

    log = []

    log_std_options = True

    vcs_control_dir = '.fake'

    revision = 'fakerev'

    name = 'fakevcs'  # for pip.vcs.VersionSupport registration

    def get_update(self, revision):
        self.revision = revision
        if not os.path.isdir(self.target_dir):
            os.mkdir(self.target_dir)
        control = os.path.join(self.target_dir, self.vcs_control_dir)
        if not os.path.isdir(control):
            os.mkdir(control)

        options = self.options.copy()
        if self.log_std_options:
            options['offline'] = self.offline
            options['clear_locks'] = self.clear_locks
        self.log.append((self.target_dir, self.url, revision, options),)

    def parents(self):
        return [self.revision]

    def archive(self, target):
        utils.mkdirp(target)
        with open(os.path.join(target, '.fake_archival.txt'), 'w') as f:
            f.write(str(self.revision))


vcs.SUPPORTED['fakevcs'] = FakeRepo
from pip.vcs import vcs as pip_vcs
pip_vcs.register(FakeRepo)  # for tests around gp.vcsdevelop


def get_vcs_log():
    return FakeRepo.log


def clear_vcs_log():
    FakeRepo.log = []


class RecipeTestCase(unittest.TestCase):
    """A base setup for tests of recipe classes"""

    def setUp(self):
        b_dir = self.buildout_dir = mkdtemp('test_oerp_base_recipe')
        eggs_dir = os.path.join(b_dir, 'eggs')
        os.mkdir(eggs_dir)
        develop_dir = os.path.join(b_dir, 'develop-eggs')
        os.mkdir(develop_dir)
        self.buildout = {}
        self.buildout['buildout'] = {
            'directory': b_dir,
            'offline': False,
            'parts-directory': os.path.join(b_dir, 'parts'),
            'bin-directory': os.path.join(b_dir, 'bin'),
            'find-links': '',
            'allow-hosts': '',
            'eggs-directory': eggs_dir,
            'develop-eggs-directory': develop_dir,
            'python': 'main_python',
        }

        self.buildout['main_python'] = dict(executable=sys.executable)

        # temporary monkey patch of easy_install to avoid actual requests to
        # PyPI (offline mode currently does not protect against that, even
        # though I checked that it is recognized by zc.recipe.egg
        # TODO this does not seem to really work in some context
        # (Debian wheezy buildslave in a virtualenv, with zc.recipe.egg 2.0.0a3
        # we see nose being downloaded several times)
        self.unreachable_distributions = set()
        Installer._orig_obtain = Installer._obtain

        def _obtain(inst, requirement, source=None):
            if requirement.project_name in self.unreachable_distributions:
                return None
            return inst._orig_obtain(requirement, source=source)
        Installer._obtain = _obtain

    def tearDown(self):
        clear_vcs_log()
        shutil.rmtree(self.buildout_dir)
        Installer._obtain = Installer._orig_obtain



import os
import logging
import subprocess
from ConfigParser import ConfigParser
from ConfigParser import NoOptionError
from .base import BaseRepo
from .base import SUBPROCESS_ENV
from .base import update_check_call
from ..utils import check_output

logger = logging.getLogger(__name__)


class HgRepo(BaseRepo):

    vcs_control_dir = '.hg'

    def update_hgrc_paths(self):
        """Update hgrc paths section if needed.

        Old paths are kept in renamed form: buildout_save_%d."""
        parser = ConfigParser()
        hgrc_path = os.path.join(self.target_dir, '.hg', 'hgrc')
        parser.read(hgrc_path)
        default = parser.get('paths', 'default')
        if default == self.url:
            return

        count = 1
        while True:
            save = 'buildout_save_%d' % count
            try:
                parser.get('paths', save)
            except NoOptionError:
                break
            count += 1

        parser.set('paths', save, default)
        parser.set('paths', 'default', self.url)
        f = open(hgrc_path, 'w')
        parser.write(f)
        f.close()

    def uncommitted_changes(self):
        """True if we have uncommitted changes."""
        return bool(check_output(['hg', '--cwd', self.target_dir, 'status'],
                                 env=SUBPROCESS_ENV))

    def parents(self):
        """Return full hash of parent nodes. """
        return check_output(['hg', '--cwd', self.target_dir, 'parents',
                             '--template={node}'],
                            env=SUBPROCESS_ENV).split()

    def have_fixed_revision(self, revstr):
        """True if revstr is a fixed revision that we already have.

        Check is done for known tags (except tip) and known nodes identified
        by a long enough (12 char) prefix of their hexadecimal hash.

        Summary of collision cases for hg up:

        - a revision number has precedence over a tag identically named
        - a tag that is a strict prefix of a full hexadecimal node hash wins
          over that node.
        - a full hexadecimal node hash wins over a tag that would be
          identically named (someone would need to be really disturbed to do
          that in real life).
        - a hexadecimal node that coincides with a decimal revision number is
          not something I can test :-)

        In theory, a 12 char hexadecimal node hash could be shadowed by an
        incoming tag. But also, any tag could be overridden. These are
        considered to be fixed anyway for convenience in sensible use-cases.

        People having CI robots involving tags that do get overridden by a
        third party upstream should complain to upstream for utterly bad
        practices.
        """
        revstr = revstr.strip()
        if revstr == 'tip' or not revstr:
            return False

        try:
            out = check_output(['hg', '--cwd', self.target_dir, 'log',
                                '-r', revstr,
                                '--template={node}\n{tags}\n{rev}'],
                               env=SUBPROCESS_ENV)
        except subprocess.CalledProcessError:
            return False

        node, tags, rev = out.split(os.linesep)

        if node.startswith(revstr):
            # if too short, can be superseded by an incoming tag
            if len(revstr) >= 12:
                logger.info("[hg] Found requested revision %r in %s",
                            revstr, self.target_dir)
                return True

        if revstr == rev:
            logger.warn("[hg] In repo %s, you should not pinpoint revision "
                        "by a local revision number such as %r",
                        self.target_dir, revstr)
            # but indeed, nothing can change it (unless one day a node has
            # exactly that hash code, chances are...)
            return True

        if revstr in tags.split():
            logger.info("[hg] In repo %s, found tag %r as %s",
                        self.target_dir, revstr, node)
            return True

        return False

    def clean(self):
        if not os.path.isdir(self.target_dir):
            return

        try:
            subprocess.check_call(['hg', 'purge', '--cwd', self.target_dir])
        except subprocess.CalledProcessError as exc:
            if exc.returncode == 255:
                # fallback to default implementation
                logger.warn("The 'purge' Mercurial extension is not "
                            "activated. Do 'hg help purge' for more details "
                            "Falling back to default cleaning implementation")
                super(HgRepo, self).clean()

    def get_update(self, revision):
        """Ensure that target_dir is a clone of url at specified revision.

        If target_dir already exists, does a simple pull.
        Offline-mode: no clone nor pull, but update.
        """
        target_dir = self.target_dir
        url = self.url
        offline = self.offline

        if not os.path.exists(target_dir):
            # TODO case of local url ?
            if offline:
                raise IOError(
                    "hg repository %r does not exist; "
                    "cannot clone it from %r (offline mode)" % (target_dir,
                                                                url))

            logger.info("Cloning %s ...", url)
            clone_cmd = ['hg', 'clone']
            if revision:
                clone_cmd.extend(['-r', revision])
            clone_cmd.extend([url, target_dir])
            subprocess.check_call(clone_cmd, env=SUBPROCESS_ENV)
        else:
            self.update_hgrc_paths()
            # TODO what if remote repo is actually local fs ?
            if self.have_fixed_revision(revision):
                self._update(revision)
                return

            if not offline:
                self._pull()
            self._update(revision)

    def _pull(self):
        logger.info("Pull for hg repo %r ...", self.target_dir)
        subprocess.check_call(['hg', '--cwd', self.target_dir, 'pull'],
                              env=SUBPROCESS_ENV)

    def _update(self, revision):
        target_dir = self.target_dir
        logger.info("Updating %s to revision %s", target_dir, revision)
        up_cmd = ['hg', '--cwd', target_dir, 'up']
        if revision:
            up_cmd.extend(['-r', revision])
        update_check_call(up_cmd, env=SUBPROCESS_ENV)

    def archive(self, target_path):
        subprocess.check_call(['hg', '--cwd', self.target_dir,
                               'archive', target_path])



# coding: utf-8
from os.path import join, basename
import os
import sys
import urllib
import tarfile
import setuptools
import logging
import stat
import imp
import shutil
import ConfigParser
import distutils.core
import pkg_resources
try:
    from collections import OrderedDict
except ImportError:  # Python < 2.7
    from ordereddict import OrderedDict  # noqa
from zc.buildout.easy_install import MissingDistribution
from zc.buildout import UserError
from zc.buildout.easy_install import VersionConflict
import zc.recipe.egg

import httplib
import rfc822
from urlparse import urlparse
from . import vcs
from . import utils

logger = logging.getLogger(__name__)


def rfc822_time(h):
    """Parse RFC 2822-formatted http header and return a time int."""
    rfc822.mktime_tz(rfc822.parsedate_tz(h))


class MainSoftware(object):
    """Placeholder to represent the main software instead of an addon location.

    Should just have a singleton instance: :data:`main_software`,
    whose meaning depends on the concrete recipe class using it.

    For example, in :class:`anybox.recipe.openerp.server.ServerRecipe`,
    :data:`main_software` represents the OpenObject server or the OpenERP
    standard distribution.
    """

    def __str__(self):
        return 'Main Software'

main_software = MainSoftware()

GP_VCS_EXTEND_DEVELOP = 'vcs-extend-develop'


class BaseRecipe(object):
    """Base class for other recipes.

    It implements notably fetching of the main software part plus addons.

    The :attr:`sources` attribute is a ``dict`` storing how to fetch the main
    software part and the specified addons, with the following structure:

       ``local path -> (type, location_spec, options)``, in which:

       :local path: is either the :data:`main_software` singleton
                    (see :class:`MainSoftware`) or a local path to an
                    addons directory.
       :type: can be either

              * ``'local'``
              * ``'downloadable'``
              * one of the supported vcs

       :location_spec: is, depending on the type, a tuple specifying how
                       fetch is to be done:

                       ``(url, None)``, or ``(vcs_url, vcs_revision)``
                       or ``None``
       :addons options: are typically used to specify that the addons
                        directory is actually a subdir of the specified one.

                        VCS support classes (see
                        :mod:`anybox.recipe.openerp.vcs`) can implemented their
                        dedicated options

    """

    default_dl_url = {'6.0': 'http://nightly.openerp.com/6.0/6.0/',
                      '6.1': 'http://nightly.openerp.com/6.1/releases/',
                      '7.0': 'http://nightly.openerp.com/7.0/releases/',
                      '5.0': 'http://nightly.openerp.com/old/openerp-5/',
                      }

    nightly_dl_url = {'6.0': 'http://nightly.openerp.com/6.0/6.0/',
                      '6.1': 'http://nightly.openerp.com/6.1/nightly/src/',
                      '7.0': 'http://nightly.openerp.com/7.0/nightly/src/',
                      'trunk': 'http://nightly.openerp.com/trunk/nightly/src/',
                      }

    recipe_requirements = ()  # distribution required for the recipe itself
    recipe_requirements_paths = ()  # a default value is useful in unit tests
    requirements = ()  # requirements for what the recipe installs to run
    soft_requirements = ()  # subset of requirements that's not necessary
    addons_paths = ()

    # Caching logic for the main OpenERP part (e.g, without addons)
    # Can be 'filename' or 'http-head'
    main_http_caching = 'filename'

    def __init__(self, buildout, name, options):
        self.requirements = list(self.requirements)
        self.recipe_requirements_path = []
        self.buildout, self.name, self.options = buildout, name, options
        self.b_options = self.buildout['buildout']
        self.buildout_dir = self.b_options['directory']
        # GR: would prefer lower() but doing as in 'zc.recipe.egg'
        # (later) the standard way for all booleans is to use
        # options.query_bool() or get_bool(), but it doesn't lower() at all
        self.offline = self.b_options['offline'] == 'true'
        self.clean = options.get('clean') == 'true'
        clear_locks = options.get('vcs-clear-locks', '').lower()
        self.vcs_clear_locks = clear_locks == 'true'
        clear_retry = options.get('vcs-clear-retry', '').lower()
        self.clear_retry = clear_retry == 'true'

        # same as in zc.recipe.eggs
        self.extra_paths = [
            join(self.buildout_dir, p.strip())
            for p in self.options.get('extra-paths', '').split(os.linesep)
            if p.strip()
        ]
        self.options['extra-paths'] = os.linesep.join(self.extra_paths)

        self.downloads_dir = self.make_absolute(
            self.b_options.get('openerp-downloads-directory', 'downloads'))
        self.version_wanted = None  # from the buildout
        self.version_detected = None  # string from the openerp setup.py
        self.parts = self.buildout['buildout']['parts-directory']
        self.openerp_dir = None
        self.archive_filename = None
        self.archive_path = None  # downloaded tar.gz

        if options.get('scripts') is None:
            options['scripts'] = ''

        # a dictionnary of messages to display in case a distribution is
        # not installable.
        self.missing_deps_instructions = {
            'PIL': ("You don't need to require it for OpenERP any more, since "
                    "the recipe automatically adds a dependency to Pillow. "
                    "If you really need it for other reasons, installing it "
                    "system-wide is a good option. "),
        }

        self.openerp_installed = []

        self.etc = self.make_absolute(options.get('etc-directory', 'etc'))
        self.bin_dir = self.buildout['buildout']['bin-directory']
        self.config_path = join(self.etc, self.name + '.cfg')
        for d in self.downloads_dir, self.etc:
            if not os.path.exists(d):
                logger.info('Created %s/ directory' % basename(d))
                os.mkdir(d)

        self.sources = OrderedDict()
        self.parse_addons(options)
        self.parse_version()
        self.parse_revisions(options)

    def parse_version(self):
        """Set the main software in :attr:`sources` and related attributes.
        """
        self.version_wanted = self.options.get('version')
        if self.version_wanted is None:
            raise ValueError('You must specify the version')

        self.preinstall_version_check()

        version_split = self.version_wanted.split()

        if len(version_split) == 1:
            # version can be a simple version name, such as 6.1-1
            major_wanted = self.version_wanted[:3]
            pattern = self.archive_filenames[major_wanted]
            if pattern is None:
                raise ValueError(
                    'OpenERP version %r'
                    'is not supported' % self.version_wanted)

            self.archive_filename = pattern % self.version_wanted
            self.archive_path = join(self.downloads_dir, self.archive_filename)
            base_url = self.options.get(
                'base_url', self.default_dl_url[major_wanted])
            self.sources[main_software] = (
                'downloadable',
                ('/'.join((base_url.strip('/'), self.archive_filename)), None))
            return

        # in all other cases, the first token is the type of version
        type_spec = version_split[0]
        if type_spec in ('local', 'path'):
            self.openerp_dir = join(self.buildout_dir, version_split[1])
            self.sources[main_software] = ('local', None)
        elif type_spec == 'url':
            url = version_split[1]
            self.archive_filename = urlparse(url).path.split('/')[-1]
            self.archive_path = join(self.downloads_dir, self.archive_filename)
            self.sources[main_software] = ('downloadable', (url, None))
        elif type_spec == 'nightly':
            if len(version_split) != 3:
                raise ValueError(
                    "Unrecognized nightly version specification: "
                    "%r (expecting series, number) % version_split[1:]")
            self.nightly_series, self.version_wanted = version_split[1:]
            type_spec = 'downloadable'
            if self.version_wanted == 'latest':
                self.main_http_caching = 'http-head'
            series = self.nightly_series
            self.archive_filename = (
                self.archive_nightly_filenames[series] % self.version_wanted)
            self.archive_path = join(self.downloads_dir, self.archive_filename)
            base_url = self.options.get('base_url',
                                        self.nightly_dl_url[series])
            self.sources[main_software] = (
                'downloadable',
                ('/'.join((base_url.strip('/'), self.archive_filename)), None))
        else:
            # VCS types
            type_spec, url, repo_dir, self.version_wanted = version_split[0:4]
            options = dict(opt.split('=') for opt in version_split[4:])
            self.openerp_dir = join(self.parts, repo_dir)
            self.sources[main_software] = (type_spec,
                                           (url, self.version_wanted), options)

    def preinstall_version_check(self):
        """Perform version checks before any attempt to install.

        To be subclassed.
        """

    def install_recipe_requirements(self):
        """Install requirements for the recipe to run."""
        to_install = self.recipe_requirements
        eggs_option = os.linesep.join(to_install)
        eggs = zc.recipe.egg.Eggs(self.buildout, '', dict(eggs=eggs_option))
        ws = eggs.install()
        _, ws = eggs.working_set()
        self.recipe_requirements_paths = [ws.by_key[dist].location
                                          for dist in to_install]
        sys.path.extend(self.recipe_requirements_paths)

    def merge_requirements(self):
        """Merge eggs option with self.requirements."""
        if 'eggs' not in self.options:
            self.options['eggs'] = '\n'.join(self.requirements)
        else:
            self.options['eggs'] += '\n' + '\n'.join(self.requirements)

    def install_requirements(self):
        """Install egg requirements and scripts.

        If some distributions are known as soft requirements, will retry
        without them
        """
        while True:
            missing = None
            eggs = zc.recipe.egg.Scripts(self.buildout, '', self.options)
            try:
                eggs.install()
            except MissingDistribution, exc:
                missing = exc.data[0].project_name
            except VersionConflict:
                raise
            except UserError, exc:  # zc.buildout >= 2.0
                missing = exc.message.split(os.linesep)[0].split()[-1]

            if missing is not None:
                msg = self.missing_deps_instructions.get(missing)
                if msg is None:
                    raise
                logger.error("Could not find %r. " + msg, missing)
                # GR this condition won't be enough in case of version
                # conditions in requirement
                if missing not in self.soft_requirements:
                    sys.exit(1)
                else:
                    attempted = self.options['eggs'].split(os.linesep)
                    self.options['eggs'] = os.linesep.join(
                        [egg for egg in attempted if egg != missing])
            else:
                break

        self.eggs_reqs, self.eggs_ws = eggs.working_set()
        self.ws = self.eggs_ws

    def apply_version_dependent_decisions(self):
        """Store some booleans depending on detected version.

        To be refined by subclasses.
        """
        pass

    @property
    def major_version(self):
        detected = self.version_detected
        if detected is None:
            return None
        return utils.major_version(detected)

    def read_release(self):
        """Try and read the release.py file directly.

        Should be used only if reading setup.py failed, which happens
        with OpenERP 5.0
        """
        with open(join(self.openerp_dir, 'bin', 'release.py'), 'rb') as f:
            mod = imp.load_module('release', f, 'release.py',
                                  ('.py', 'r', imp.PY_SOURCE))
        self.version_detected = mod.version

    def read_openerp_setup(self):
        """Ugly method to extract requirements & version from ugly setup.py.

        Primarily designed for 6.0, but works with 6.1 as well.
        """
        old_setup = setuptools.setup
        old_distutils_setup = distutils.core.setup  # 5.0 directly imports this

        def new_setup(*args, **kw):
            self.requirements.extend(kw.get('install_requires', ()))
            self.version_detected = kw['version']
        setuptools.setup = new_setup
        distutils.core.setup = new_setup
        sys.path.insert(0, '.')
        with open(join(self.openerp_dir, 'setup.py'), 'rb') as f:
            saved_argv = sys.argv
            sys.argv = ['setup.py', 'develop']
            try:
                imp.load_module('setup', f, 'setup.py',
                                ('.py', 'r', imp.PY_SOURCE))
            except SystemExit as exception:
                msg = exception.message
                if not isinstance(msg, int) and 'dsextras' in msg:
                    raise EnvironmentError(
                        'Please first install PyGObject and PyGTK !')
                else:
                    try:
                        self.read_release()
                    except Exception as exc:
                        raise EnvironmentError(
                            'Problem while reading OpenERP release.py: '
                            + exc.message)
            except ImportError, exception:
                if 'babel' in exception.message:
                    raise EnvironmentError(
                        'OpenERP setup.py has an unwanted import Babel.\n'
                        '=> First install Babel on your system or '
                        'virtualenv :(\n'
                        '(sudo aptitude install python-babel, '
                        'or pip install babel)')
                else:
                    raise exception
            except Exception, exception:
                raise EnvironmentError('Problem while reading OpenERP '
                                       'setup.py: ' + exception.message)
            finally:
                sys.argv = saved_argv
        sys.path.pop(0)
        setuptools.setup = old_setup
        distutils.core.setup = old_distutils_setup
        self.apply_version_dependent_decisions()

    def make_absolute(self, path):
        """Make a path absolute if needed.

        If not already absolute, it is interpreted as relative to the
        buildout directory."""
        if os.path.isabs(path):
            return path
        return join(self.buildout_dir, path)

    def sandboxed_tar_extract(self, sandbox, tarfile, first=None):
        """Extract those members that are below the tarfile path 'sandbox'.

        The tarfile module official doc warns against attacks with .. in tar.

        The option to start with a first member is useful for this case, since
        the recipe consumes a first member in the tar file to get the openerp
        main directory in parts.
        It is taken for granted that this first member has already been
        checked.
        """

        if first is not None:
            tarfile.extract(first)

        for tinfo in tarfile:
            if tinfo.name.startswith(sandbox + '/'):
                tarfile.extract(tinfo)
            else:
                logger.warn('Tarball member %r is outside of %r. Ignored.',
                            tinfo, sandbox)

    def _produce_setup_without_pil(self, src_directory):
        """Create a copy of setup.py without PIL and return a path to it."""

        new_setup_path = join(src_directory, 'setup.nopil.py')
        with open(join(src_directory, 'setup.py')) as inp:
            setup_str = inp.read()
        with open(new_setup_path, 'w') as out:
            out.write(setup_str.replace("'PIL',", ''))
        return new_setup_path

    def develop(self, src_directory, setup_has_pil=False):
        """Develop the specified source distribution.

        Any call to zc.recipe.eggs will use that developped version.
        develop() launches a subprocess, to which we need to forward
        the paths to requirements via PYTHONPATH.
        If setup_has_pil is True, an altered version of setup that does not
        require it is produced to perform the develop.
        """
        logger.debug("Developing %r", src_directory)
        develop_dir = self.b_options['develop-eggs-directory']
        pythonpath_bak = os.getenv('PYTHONPATH')
        os.putenv('PYTHONPATH', ':'.join(self.recipe_requirements_paths))

        if setup_has_pil:
            setup = self._produce_setup_without_pil(src_directory)
        else:
            setup = src_directory

        try:
            zc.buildout.easy_install.develop(setup, develop_dir)
        finally:
            if setup_has_pil:
                os.unlink(setup)

        if pythonpath_bak is None:
            os.unsetenv('PYTHONPATH')
        else:
            os.putenv('PYTHONPATH', pythonpath_bak)

    def parse_addons(self, options):
        """Parse the addons options into :attr:`sources`.

        See :class:`BaseRecipe` for the structure of :attr:`sources`.
        """

        for line in options.get('addons', '').split(os.linesep):
            split = line.split()
            if not split:
                return
            loc_type = split[0]
            spec_len = 2 if loc_type == 'local' else 4

            options = dict(opt.split('=') for opt in split[spec_len:])
            if loc_type == 'local':
                addons_dir = split[1]
                location_spec = None
            else:  # vcs
                repo_url, addons_dir, repo_rev = split[1:4]
                location_spec = (repo_url, repo_rev)

            addons_dir = addons_dir.rstrip('/')  # trailing / can be harmful
            self.sources[addons_dir] = (loc_type, location_spec, options)

    def parse_revisions(self, options):
        """Parse revisions options and update :attr:`sources`.

        It is assumed that :attr:`sources` has already been populated, and
        notably has a :data:`main_software` entry.
        This allows for easy fixing of revisions in an extension buildout

        See :class:`BaseRecipe` for the structure of :attr:`sources`.
        """
        for line in options.get('revisions', '').split(os.linesep):
            # GR inline comment should have not gone through, but sometimes
            # does (see lp:1130590). This below does not exactly conform to
            # spec http://docs.python.org/2/library/configparser.html
            # (we don't check for whitespace before separator), but is good
            # enough in this case.
            line = line.split(';', 1)[0].strip()
            if not line:
                continue

            split = line.split()
            if len(split) > 2:
                raise ValueError("Invalid revisions line: %r" % line)

            # addon or main software
            if len(split) == 2:
                local_path = split[0]
            else:
                local_path = main_software
            revision = split[-1]

            source = self.sources.get(local_path)
            if source is None:  # considered harmless for now
                logger.warn("Ignoring attempt to fix revision on unknown "
                            "source %r. You may have a leftover to clean",
                            local_path)
                continue

            if source[0] in ('downloadable', 'local'):
                raise ValueError("In revision line %r : can't fix a revision "
                                 "for non-vcs source" % line)

            logger.info("%s will be on revision %r", local_path, revision)
            self.sources[local_path] = ((source[0], (source[1][0], revision))
                                        + source[2:])

    def retrieve_addons(self):
        """Peform all lookup and downloads specified in :attr:`sources`.

        See :class:`BaseRecipe` for the structure of :attr:`sources`.
        """
        self.addons_paths = []
        for local_dir, source_spec in self.sources.items():
            if local_dir is main_software:
                continue

            loc_type, loc_spec, addons_options = source_spec
            local_dir = self.make_absolute(local_dir)
            options = dict(offline=self.offline,
                           clear_locks=self.vcs_clear_locks,
                           clean=self.clean)
            options.update(addons_options)

            if loc_type != 'local':
                for k, v in self.options.items():
                    if k.startswith(loc_type + '-'):
                        options[k] = v

                repo_url, repo_rev = loc_spec
                vcs.get_update(loc_type, local_dir, repo_url, repo_rev,
                               clear_retry=self.clear_retry,
                               **options)
            elif self.clean:
                utils.clean_object_files(local_dir)

            subdir = addons_options.get('subdir')
            addons_dir = join(local_dir, subdir) if subdir else local_dir

            manifest = os.path.join(addons_dir, '__openerp__.py')
            manifest_pre_v6 = os.path.join(addons_dir, '__terp__.py')
            if os.path.isfile(manifest) or os.path.isfile(manifest_pre_v6):
                if loc_type == 'local':
                    raise ValueError(
                        "Local addons line %r should refer to a directory "
                        "containing addons, not to a standalone addon. "
                        "The recipe can perform automatic creation of "
                        "intermediate directories for VCS cases only"
                        % addons_dir)
                # repo is a single addon, put it actually below
                name = os.path.split(addons_dir)[1]
                c = 0
                tmp = addons_dir + '_%d' % c
                while os.path.exists(tmp):
                    c += 1
                    tmp = addons_dir + '_%d' % c
                os.rename(addons_dir, tmp)
                os.mkdir(addons_dir)
                new_dir = join(addons_dir, name)
                os.rename(tmp, new_dir)
            self.addons_paths.append(addons_dir)

    def main_download(self):
        """HTTP download for main part of the software to self.archive_path.
        """
        if self.offline:
            raise IOError("%s not found, and offline "
                          "mode requested" % self.archive_path)
        url = self.sources[main_software][1][0]
        logger.info("Downloading %s ..." % url)

        try:
            msg = urllib.urlretrieve(url, self.archive_path)
            if msg[1].type == 'text/html':
                os.unlink(self.archive_path)
                raise LookupError(
                    'Wanted version %r not found on server (tried %s)' % (
                        self.version_wanted, url))

        except (tarfile.TarError, IOError):
            # GR: ContentTooShortError subclasses IOError
            os.unlink(self.archive_path)
            raise IOError('The archive does not seem valid: ' +
                          repr(self.archive_path))

    def is_stale_http_head(self):
        """Tell if the download is stale by doing a HEAD request.

        Assumes the correct date had been written upon download.
        This is the same system as in GNU Wget 1.12. It works even if
        the server does not implement conditional responses such as 304
        """
        archivestat = os.stat(self.archive_path)
        length, modified = archivestat.st_size, archivestat.st_mtime

        url = self.sources[main_software][1][0]
        logger.info("Checking if %s if fresh wrt %s",
                    self.archive_path, url)
        parsed = urlparse(url)
        if parsed.scheme == 'https':
            cnx_cls = httplib.HTTPSConnection
        else:
            cnx_cls = httplib.HTTPConnection
        try:
            cnx = cnx_cls(parsed.netloc)
            cnx.request('HEAD', parsed.path)  # TODO query ? fragment ?
            res = cnx.getresponse()
        except IOError:
            return True

        if res.status != 200:
            return True

        if int(res.getheader('Content-Length')) != length:
            return True

        head_modified = res.getheader('Last-Modified')
        logger.debug("Last-modified from HEAD request: %s", head_modified)
        if rfc822_time(head_modified) > modified:
            return True

        logger.info("No need to re-download %s", self.archive_path)

    def retrieve_main_software(self):
        """Lookup or fetch the main software.

        See :class:`MainSoftware` and :class:`BaseRecipe` for explanations.
        """

        source = self.sources[main_software]
        type_spec = source[0]
        logger.info('Selected install type: %s', type_spec)
        if type_spec == 'local':
            logger.info('Local directory chosen, nothing to do')
            if self.clean:
                utils.clean_object_files(self.openerp_dir)
        elif type_spec == 'downloadable':
            # download if needed
            if ((self.archive_path and not os.path.exists(self.archive_path))
                or (self.main_http_caching == 'http-head'
                    and self.is_stale_http_head())):
                self.main_download()

            logger.info(u'Inspecting %s ...' % self.archive_path)
            tar = tarfile.open(self.archive_path)
            first = tar.next()
            # Everything that follows assumes all tarball members
            # are inside a directory with an expected name such
            # as openerp-6.1-1
            assert(first.isdir())
            extracted_name = first.name.split('/')[0]
            self.openerp_dir = join(self.parts, extracted_name)
            # protection against malicious tarballs
            assert(not os.path.isabs(extracted_name))
            assert(self.openerp_dir.startswith(self.parts))

            logger.info("Cleaning existing %s", self.openerp_dir)
            if os.path.exists(self.openerp_dir):
                shutil.rmtree(self.openerp_dir)
            logger.info(u'Extracting %s ...' % self.archive_path)
            self.sandboxed_tar_extract(extracted_name, tar, first=first)
            tar.close()
        else:
            url, rev = source[1]
            options = dict((k, v) for k, v in self.options.iteritems()
                           if k.startswith(type_spec + '-'))
            options.update(source[2])
            if self.clean:
                options['clean'] = True
            vcs.get_update(type_spec, self.openerp_dir, url, rev,
                           offline=self.offline,
                           clear_retry=self.clear_retry, **options)

    def _register_extra_paths(self):
        """Add openerp paths into the extra-paths (used in scripts' sys.path).
        """
        extra = self.extra_paths
        if self.major_version >= (6, 2):
            # TODO still necessary ?
            extra.extend((self.openerp_dir,
                         join(self.openerp_dir, 'addons')))
        else:
            extra.extend((join(self.openerp_dir, 'bin'),
                          join(self.openerp_dir, 'bin', 'addons')))
        self.options['extra-paths'] = os.linesep.join(extra)

    def install(self):
        os.chdir(self.parts)

        freeze_to = self.options.get('freeze-to')
        extract_downloads_to = self.options.get('extract-downloads-to')

        if ((freeze_to is not None or extract_downloads_to is not None)
                and not self.offline):
            raise ValueError("To freeze a part, you must run offline "
                             "so that there's no modification from what "
                             "you just tested. Please rerun with -o.")

        if extract_downloads_to is not None and freeze_to is None:
            freeze_to = os.path.join(extract_downloads_to,
                                     'extracted_from.cfg')

        self.retrieve_main_software()
        self.retrieve_addons()

        self.install_recipe_requirements()
        os.chdir(self.openerp_dir)  # GR probably not needed any more
        self.read_openerp_setup()

        if (self.sources[main_software][0] == 'downloadable'
                and self.version_wanted == 'latest'):
            self.nightly_version = self.version_detected.split('-', 1)[1]
            logger.warn("Detected 'nightly latest version', you may want to "
                        "fix it in your config file for replayability: \n    "
                        "version = " + self.dump_nightly_latest_version())

        self.finalize_addons_paths()
        self._register_extra_paths()

        if self.version_detected is None:
            raise EnvironmentError('Version of OpenERP could not be detected')
        self.merge_requirements()
        self.install_requirements()

        self._install_startup_scripts()

        # create the config file
        if os.path.exists(self.config_path):
            os.remove(self.config_path)
        logger.info('Creating config file: %s',
                    os.path.relpath(self.config_path, self.buildout_dir))
        self._create_default_config()

        # modify the config file according to recipe options
        config = ConfigParser.RawConfigParser()
        config.read(self.config_path)
        for recipe_option in self.options:
            if '.' not in recipe_option:
                continue
            section, option = recipe_option.split('.', 1)
            if not config.has_section(section):
                config.add_section(section)
            config.set(section, option, self.options[recipe_option])
        with open(self.config_path, 'wb') as configfile:
            config.write(configfile)

        if extract_downloads_to:
            self.extract_downloads_to(extract_downloads_to)
        if freeze_to:
            self.freeze_to(freeze_to)
        return self.openerp_installed

    def dump_nightly_latest_version(self):
        """After download/analysis of 'nightly latest', give equivalent spec.
        """
        return ' '.join(('nightly', self.nightly_series, self.nightly_version))

    def freeze_to(self, out_config_path):
        """Create an extension buildout freezing current revisions & versions.
        """

        logger.info("Freezing part %r to config file %r", self.name,
                    out_config_path)
        out_conf = ConfigParser.ConfigParser()

        frozen = getattr(self.buildout, '_openerp_recipe_frozen', None)
        if frozen is None:
            frozen = self.buildout._openerp_recipe_frozen = set()

        if out_config_path in frozen:
            # read configuration started by other recipe
            out_conf.read(self.make_absolute(out_config_path))
        else:
            self._prepare_frozen_buildout(out_conf)

        self._freeze_egg_versions(out_conf, 'versions')

        out_conf.add_section(self.name)
        addons_option = []
        self.local_modifications = []
        for local_path, source in self.sources.items():
            source_type = source[0]
            if source_type == 'local':
                continue

            if local_path is main_software:
                if source_type == 'downloadable':
                    self._freeze_downloadable_main_software(out_conf)
                else:  # vcs
                    abspath = self.openerp_dir
                    self.cleanup_openerp_dir()
            else:
                abspath = self.make_absolute(local_path)

            if source_type == 'downloadable':
                continue

            revision = self._freeze_vcs_source(source_type, abspath)
            if local_path is main_software:
                addons_option.insert(0, '%s  ; main software part' % revision)
                # actually, that comment will be lost if this is not the
                # last part (dropped upon reread)
            else:
                addons_option.append(' '.join((local_path, revision)))

        if addons_option:
            out_conf.set(self.name, 'revisions',
                         os.linesep.join(addons_option))
        if self.local_modifications:

            logger.error(
                "Uncommitted changes and/or untracked files in: %s"
                "Unsafe to freeze. Please commit or revert and test again !",
                os.linesep.join(
                    ['', ''] + ['   - ' + p
                                for p in self.local_modifications] + ['', '']))

            sys.exit(17)  # GR I like that number

        with open(self.make_absolute(out_config_path), 'w') as out:
            out_conf.write(out)
        frozen.add(out_config_path)

    def _get_gp_vcs_develops(self):
        """Return a tuple of (raw, parsed) vcs-extends-develop specifications.
        """
        lines = self.b_options.get(
            GP_VCS_EXTEND_DEVELOP)
        if not lines:
            return ()

        try:
            import pip.req
        except ImportError:
            logger.error("You have vcs-extends-develop distributions "
                         "but pip is not available. That means that "
                         "gp.vcsdevelop is not properly installed. Did "
                         "you ever run that buildout ?")
            raise

        return tuple((line, pip.req.parse_editable(line))
                     for line in lines.split(os.linesep) if line)

    def _prepare_frozen_buildout(self, conf):
        """Create the 'buildout' section in conf."""
        conf.add_section('buildout')
        conf.set('buildout', 'extends', self.buildout_cfg_name())
        conf.add_section('versions')
        conf.set('buildout', 'versions', 'versions')

        # freezing for gp.vcsdevelop
        extends = []
        for raw, parsed in self._get_gp_vcs_develops():
            local_path = parsed[0]
            hash_split = raw.rsplit('#')
            url = hash_split[0]
            url = url.rsplit('@', 1)[0]
            vcs_type = url.split('+', 1)[0]
            # vcs-develop process adds .egg-info file (often forgotten in VCS
            # ignore files) and changes setup.cfg.
            # For now we'll have to allow local modifications.
            revision = self._freeze_vcs_source(vcs_type,
                                               self.make_absolute(local_path),
                                               allow_local_modification=True)
            extends.append('%s@%s#%s' % (url, revision, hash_split[1]))

        conf.set('buildout', GP_VCS_EXTEND_DEVELOP, os.linesep.join(extends))

    def _freeze_downloadable_main_software(self, conf):
        """If needed, sets the main version option in ConfigParser.

        Currently does not dump the fully resolved URL, since future
        reproduction may be better done with another URL base holding archived
        old versions : it's better to let tomorrow logic handle that
        from higher level information.
        """

        if self.version_wanted == 'latest':
            conf.set(self.name, 'version', self.dump_nightly_latest_version())

    def _freeze_egg_versions(self, conf, section, exclude=()):
        """Update a ConfigParser section with current working set egg versions.
        """
        versions = dict((name, conf.get(section, name))
                        for name in conf.options(section))
        versions.update((name, egg.version)
                        for name, egg in self.ws.by_key.items()
                        if name not in exclude
                        and egg.precedence != pkg_resources.DEVELOP_DIST
                        )
        for name, version in versions.items():
            conf.set(section, name, version)

        # forbidding picked versions if this zc.buildout supports it right away
        # i.e. we are on zc.buildout >= 2.0
        allow_picked = self.options.get('freeze-allow-picked-versions', '')
        if allow_picked.strip() == 'false':
            pick_opt = 'allow-picked-versions'
            if pick_opt in self.b_options:
                conf.set('buildout', pick_opt, 'false')

    def _freeze_vcs_source(self, vcs_type, abspath,
                           allow_local_modification=False):
        """Return the current revision for that VCS source."""

        repo_cls = vcs.SUPPORTED[vcs_type]
        abspath = repo_cls.fix_target(abspath)
        repo = repo_cls(abspath, '')  # no need of remote URL

        if not allow_local_modification and repo.uncommitted_changes():
            self.local_modifications.append(abspath)

        parents = repo.parents()
        if len(parents) > 1:
            self.local_modifications.append(abspath)

        return parents[0]

    def extract_downloads_to(self, target_dir, outconf_name='release.cfg'):
        """Extract anything that has been downloaded to target_dir.

        This doesn't copy intermediary buildout configurations nor local parts.
        In the purpose of making a self-contained and offline playable archive,
        these are assumed to be already taken care of.
        """
        logger.info("Extracting part %r to directory %r and config file %r "
                    "therein.", self.name, target_dir, outconf_name)
        target_dir = self.make_absolute(target_dir)
        out_conf = ConfigParser.ConfigParser()

        all_extracted = getattr(self.buildout, '_openerp_recipe_extracted',
                                None)
        if all_extracted is None:
            all_extracted = self.buildout._openerp_recipe_extracted = {}
        out_config_path = join(target_dir, outconf_name)

        # GR TODO this will fail if same target dir has been used with
        # a different outconf_name
        if target_dir in all_extracted:
            # read configuration started by other recipe
            out_conf.read(out_config_path)
            extracted = all_extracted[target_dir]
        else:
            self._prepare_extracted_buildout(out_conf, target_dir)
            extracted = all_extracted[target_dir] = set()

        self._freeze_egg_versions(out_conf, 'versions')
        self._extract_sources(out_conf, target_dir, extracted)
        with open(out_config_path, 'w') as out:
            out_conf.write(out)

    def _extract_sources(self, out_conf, target_dir, extracted):
        """Core extraction method.

        out_conf is a ConfigParser instance to write to
        extracted is a technical set used to know what targets have already
        been written by previous parts and store for subsequent ones.
        """

        if not os.path.exists(target_dir):
            os.mkdir(target_dir)

        out_conf.add_section(self.name)
        addons_option = []
        for local_path, source in self.sources.items():
            source_type = source[0]
            if local_path is main_software:
                rel_path = self._extract_main_software(source_type, target_dir,
                                                       extracted)
                out_conf.set(self.name, 'version', 'local ' + rel_path)
                continue

            addons_line = ['local', local_path]
            addons_line.extend('%s=%s' % (opt, val)
                               for opt, val in source[2].items())
            addons_option.append(' '.join(addons_line))

            abspath = self.make_absolute(local_path)
            if source_type == 'downloadable':
                shutil.copytree(abspath,
                                os.path.join(target_dir, local_path))
            elif source_type != 'local':  # vcs
                self._extract_vcs_source(source_type, abspath, target_dir,
                                         local_path, extracted)

        out_conf.set(self.name, 'addons', os.linesep.join(addons_option))
        if self.options.get('revisions'):
            out_conf.set(self.name, 'revisions', '')
            # GR hacky way to make a comment for a void value. Indeed,
            # "revisions = ; comment" is not recognized as an inline comment
            # because of overall stripping and a need for whitespace before
            # the semicolon (sigh)
            out_conf.set(self.name, '; about revisions',
                         "the extended buildout '%s' uses the 'revisions' "
                         "option. The present override disables it "
                         "because it makes no sense after extraction and "
                         "replacement by the "
                         "'local' scheme" % self.buildout_cfg_name())

    def _extract_vcs_source(self, vcs_type, repo_path, target_dir, local_path,
                            extracted):
        """Extract a VCS source.

        The extracted argument is a set of previously extracted targets.
        This is because some VCS will refuse an empty directory (bzr does)
        """

        repo_path = self.make_absolute(repo_path)
        repo_cls = vcs.SUPPORTED[vcs_type]
        fixed_repo_path = repo_cls.fix_target(repo_path)
        target_path = os.path.join(target_dir, local_path)
        if fixed_repo_path != repo_path:
            # again the problem of shifts, temp lame solution
            split = os.path.split(fixed_repo_path)
            assert split[0] == repo_path
            target_path = os.path.join(target_path, split[1])

        utils.mkdirp(target_path)
        if target_path in extracted:
            return

        repo = repo_cls(fixed_repo_path, '')  # no need of remote URL
        repo.archive(target_path)
        extracted.add(target_path)

    def _extract_main_software(self, source_type, target_dir, extracted):
        """Extract the main software to target_dir and return relative path.

        As this is for extract_downloads_to, local main software is not
        extracted (supposed to be taken care of by the tool that does the
        archival of buildout dir itself).

        The extracted set avoids extracting twice to same target (refused
        by some VCSes anyway)
        """
        if not self.openerp_dir.startswith(self.buildout_dir):
            raise RuntimeError(
                "Main openerp directory %r outside of buildout "
                "directory, don't know how to handle that" % self.openerp_dir)

        local_path = self.openerp_dir[len(self.buildout_dir + os.sep):]
        target_path = join(target_dir, local_path)
        if target_path in extracted:
            return local_path

        if source_type == 'downloadable':
            shutil.copytree(self.openerp_dir, target_path)
        elif source_type != 'local':  # see docstring for 'local'
            self._extract_vcs_source(source_type, self.openerp_dir, target_dir,
                                     local_path, extracted)
        return local_path

    def _prepare_extracted_buildout(self, conf, target_dir):
        """Create the 'buildout' section in ``conf``.

        Also takes care of gp.vcsdevelop driven distributions.

        In most cases, at this stage, gp.vcsdevelop and regular develop
        distributions are expressed with absolute paths. This method will make
        them local in the destination ``conf``

        Regular develop distributions pointing outside of buildout directory
        are kept as is, assuming this has been specified in absolute form
        in the config file, hence to some resources that are outside of
        the recipe control, that are therefore expected to be deployed before
        hand on target systems.
        """
        conf.add_section('buildout')
        conf.set('buildout', 'extends', self.buildout_cfg_name())
        conf.add_section('versions')
        conf.set('buildout', 'versions', 'versions')

        develops = set(self.b_options.get('develop', '').split(os.linesep))

        extracted = set()
        for raw, parsed in self._get_gp_vcs_develops():
            local_path = parsed[0]
            abs_path = self.make_absolute(local_path)
            vcs_type = raw.split('+', 1)[0]
            self._extract_vcs_source(vcs_type, abs_path,
                                     target_dir, local_path, extracted)
            develops.add(abs_path)  # looks silly, but better for uniformity

        bdir = os.path.join(self.buildout_dir, '')
        conf.set('buildout', 'develop',
                 os.linesep.join(d[len(bdir):] if d.startswith(bdir) else d
                                 for d in develops))
        conf.set('buildout', GP_VCS_EXTEND_DEVELOP, '')

    def _install_script(self, name, content):
        """Install and register a script with prescribed name and content.

        Return the script path
        """
        path = join(self.bin_dir, name)
        f = open(path, 'w')
        f.write(content)
        f.close()
        os.chmod(path, stat.S_IRWXU)
        self.openerp_installed.append(path)
        return path

    def _install_startup_scripts(self):
        raise NotImplementedError

    def _create_default_config(self):
        raise NotImplementedError

    update = install

    def _60_fix_root_path(self):
        """Correction of root path for OpenERP 6.0 pure python install

        Actual implementation is up to subclasses
        """

    def _60_default_addons_path(self):
        """Set the default addons path for OpenERP 6.0 pure python install

        Actual implementation is up to subclasses
        """

    def _default_addons_path(self):
        """Set the default addons path for OpenERP > 6.0 pure python install

        Actual implementation is up to subclasses
        """

    def finalize_addons_paths(self, check_existence=True):
        """Add implicit paths and serialize in the addons_path option.

        if check_existence is True, all the paths will be checked for
        existence.
        """
        opt_key = 'options.addons_path'
        if opt_key in self.options:
            raise UserError("In part %r, direct use of %s is prohibited. "
                            "please use addons lines with type 'local' "
                            "instead." % (self.name, opt_key))

        if self.major_version <= (6, 0):
            base_addons = join(self.openerp_dir, 'bin', 'addons')
        else:
            base_addons = join(self.openerp_dir, 'openerp', 'addons')
        if os.path.exists(base_addons):
            self.addons_paths.insert(0, base_addons)

        if check_existence:
            for path in self.addons_paths:
                assert os.path.isdir(path), (
                    "Not a directory: %r (aborting)" % path)

        self.options['options.addons_path'] = ','.join(self.addons_paths)
        if self.major_version <= (6, 0):
            self._60_fix_root_path()

    def cleanup_openerp_dir(self):
        """Revert local modifications that have been made during installation.

        These can be, e.g., forbidden by the freeze process."""

        shutil.rmtree(join(self.openerp_dir, 'openerp.egg-info'))
        # setup rewritten without PIL is cleaned during the process itself

    def buildout_cfg_name(self, argv=None):
        """Return the name of the config file that's been called.
        """

        # not using optparse because it's not obvious how to tell it to
        # consider just one option and ignore the others.

        if argv is None:
            argv = sys.argv[1:]

        # -c FILE or --config FILE syntax
        for opt in ('-c', '--config'):
            try:
                i = argv.index(opt)
            except ValueError:
                continue
            else:
                return argv[i+1]

        # --config=FILE syntax
        prefix = "--config="
        for a in argv:
            if a.startswith(prefix):
                return a[len(prefix):]

        return 'buildout.cfg'



from unittest import TestCase
from .. import already_imported, clear_import_registry


class TestImportRegistry(TestCase):

    def tearDown(self):
        clear_import_registry()

    def test_clear(self):
        self.assertFalse(already_imported('myaddon'))
        self.assertTrue(already_imported('myaddon'))
        clear_import_registry()
        self.assertFalse(already_imported('myaddon'))

    def test_already_imported(self):
        self.assertFalse(already_imported('myaddon'))
        self.assertTrue(already_imported('myaddon'))
        self.assertTrue(already_imported('openerp.addons.myaddon'))



"""Utilities to start a server process."""
import warnings
import sys
import os
import logging
from distutils.version import Version


try:
    import openerp
except ImportError:
    warnings.warn("This must be imported with a buildout openerp recipe "
                  "driven sys.path", RuntimeWarning)
else:
    try:
        from openerp.cli import server as startup
    except ImportError:
        from .backports.cli import server as startup
    from openerp.tools import config
    from openerp import SUPERUSER_ID
    from openerp.tools.parse_version import parse_version

from optparse import OptionParser  # we support python >= 2.6

logger = logging.getLogger(__name__)

DEFAULT_VERSION_PARAMETER = 'buildout.db_version'

DEFAULT_VERSION_FILE = 'VERSION.txt'


class OpenERPVersion(Version):
    """OpenERP idea of version, wrapped in a class.

    This is based on :meth:`openerp.tools.parse_version`, and
    Provides straight-ahead comparison with tuples of integers, or
    distutils Version classes.
    """

    def parse(self, incoming):
        if isinstance(incoming, OpenERPVersion):
            self.vstring = incoming.vstring
            self.components = incoming.components
        else:
            self.vstring = incoming
            self.components = parse_version(incoming)

    def __str__(self):
        return self.vstring

    def __repr__(self):
        return 'OpenERPVersion(%r)' % str(self)

    def __cmp__(self, other):
        if isinstance(other, tuple):
            other = '.'.join(str(s) for s in other)
        elif not isinstance(other, self.__class__):
            other = str(other)  # Works with distutils' Version classes

        other = self.__class__(other)
        return cmp(self.components, other.components)


class Session(object):
    """A class to give server-level access to one database.

    There should be exactly one instance of this class per process.
    It can be used for any kind of script involving OpenERP API, and provides
    facilities for upgrade scripts (see also
    :mod:anybox.recipe.openerp.runtime.upgrade)

    Before actual use, call :meth:`open`.
    Then you'll have useful attributes and methods reminiscent of the unit test
    classes:

    * :attr:`cr`: a cursor
    * :attr:`uid`: user id
    * :attr:`registry`: access to model objects
    * :attr:`is_initialization`: True if and only if the database was not
      initialized before the call to :meth:`open`

    Example application code::

       session.open(db_name="my_db")
       admin = session.registry('res_users').browse(session.cr, session.uid, 1)
       (...)
       session.cr.commit()
       session.close()

    Transaction management is up to user code

    Upgrade scripts writers should check the version handling properties:

    * :meth:`db_version`
    * :meth:`package_version`

    Instantiation is done by passing the path to OpenERP main
    configuration file and the path of the buildout directory.

    Usually, instantiation code is written by the recipe in the body of the
    executable "OpenERP scripts" it produces.
    Script writers provide a callable that takes a
    :class:`.Session` object argument and declare it as a console script entry
    point in their distribution.
    End users can reference such entry points in their buildout configurations
    to have buildout produce the actual executable. See :doc:`/scripts`
    for details.

    Upgrade scripts are a special case of that process, in which the entry
    point is actually provided by the recipe and rewraps a user-level
    source script.

    Later versions of the recipe may find a way to pass the whole buildout
    configuration (recall that this is to be used in a separate process in
    which the buildout configuration has not been parsed).
    """

    def __init__(self, conffile, buildout_dir, parse_config=True):
        self.buildout_dir = buildout_dir
        self.openerp_config_file = conffile

        self._registry = self.cr = None
        if parse_config:
            config.parse_config(['-c', conffile])

    def ready(self):
        return self._registry is not None

    def open(self, db=None, with_demo=False):
        """Load the database

        Loading an empty database in OpenERP has the side effect of installing
        the ``base`` module. Whether to loading demo data or not has therefore
        to be decided right away.

        :param db: database name. If not specified, the same cascading of
                   defaults as OpenERP mainstream will be applied:
                   configuration file, psycopg2/lipq defaults.
        :param with_demo: controls the loading of demo data for all
                          module installations triggered by this call to
                          :meth:`open` and further uses of :meth:`load_modules`
                          on this :class:`Session` instance:

                          * if ``True``, demo data will uniformly be loaded
                          * if ``False``, no demo data will be loaded
                          * if ``None``, demo data will be loaded according to
                            the value of ``without_demo`` in configuration

                          In all cases, the behaviour will stay consistent
                          until the next call of ``open()``, but the
                          implementation does not protect against any race
                          conditions in OpenERP internals.
        """
        if db is None:
            db = config['db_name']
        if not db:
            db = ''  # expected value expected by OpenERP to start defaulting.

        cnx = openerp.sql_db.db_connect(db)
        cr = cnx.cursor()
        self.is_initialization = not(openerp.modules.db.is_initialized(cr))
        cr.close()

        startup.check_root_user()
        startup.check_postgres_user()
        openerp.netsvc.init_logger()

        saved_without_demo = config['without_demo']
        if with_demo is None:
            with_demo = config['without_demo']

        config['without_demo'] = not with_demo
        self.with_demo = with_demo

        self._registry = openerp.modules.registry.RegistryManager.get(
            db, update_module=False)
        config['without_demo'] = saved_without_demo
        self.init_cursor()
        self.uid = SUPERUSER_ID

    # A later version might read that from buildout configuration.
    _version_parameter_name = DEFAULT_VERSION_PARAMETER

    @property
    def version_file_path(self):
        """Absolute path of the flat file storing the package version.

        For now this is not configurable, a later version might read it
        from buildout configuration.
        """
        return os.path.join(self.buildout_dir, DEFAULT_VERSION_FILE)

    def parse_version_string(self, vstring):
        """Stable method for downstream code needing to instantiate a version.

        This method returns an appropriate version instance, without
        any dependency on where to import the class from. Especially useful
        for applications whose life started before this set of utilities has
        been used : this helps building an usable default.
        """
        return OpenERPVersion(vstring)

    @property
    def db_version(self):
        """Settable property for version stored in DB of the whole buildout.

        This can be thought as the latest version to which the DB has been
        upgraded to.
        A simple caching system to avoid querying the DB multiple times is
        implemented.
        """
        db_version = getattr(self, '_db_version', None)
        if db_version is not None:
            return db_version

        db_version = self.registry('ir.config_parameter').get_param(
            self.cr, self.uid, self._version_parameter_name)
        if not db_version:
            # as usual OpenERP thinks its simpler to use False as None
            # restoring sanity ASAP
            db_version = None
        else:
            db_version = OpenERPVersion(db_version)
        self._db_version = db_version
        return db_version

    @db_version.setter
    def db_version(self, version):
        self.registry('ir.config_parameter').set_param(
            self.cr, self.uid, self._version_parameter_name, str(version))
        self._db_version = OpenERPVersion(version)

    @property
    def package_version(self):
        """Property reading the version file from buildout directory.

        Comments introduced with a hash are accepted.
        Only the first significant line is taken into account.
        """
        pkg_version = getattr(self, '_pkg_version', None)
        if pkg_version is not None:
            return pkg_version

        try:
            with open(self.version_file_path) as f:
                for line in f:
                    line = line.split('#', 1)[0].strip()
                    if not line:
                        continue
                    self._pkg_version = OpenERPVersion(line)
                    return self._pkg_version
        except IOError:
            logger.info("No version file could be read, "
                        "package version considered to be None")

    def update_modules_list(self):
        """Update the list of available OpenERP modules, like the UI allows to.

        This is necessary prior to install of any new module.
        """
        self.registry('ir.module.module').update_list(self.cr, self.uid)

    def init_cursor(self):
        self.cr = self._registry.db.cursor()

    def registry(self, model):
        """Lookup model by name and return a ready-to-work instance."""
        return self._registry.get(model)

    def rollback(self):
        self.cr.rollback()

    def close(self):
        """Close the cursor and forget about the current database.

        The session is thus ready to open another database.
        """
        dbname = self.cr.dbname
        self.cr.close()
        openerp.modules.registry.RegistryManager.delete(dbname)

    def update_modules(self, modules, db=None):
        """Update the prescribed modules in the database.

        :param db: Database name. If not specified, it is assumed to have
                   already been opened with :meth:`open`, e.g, for a prior
                   read of :meth:`db_version`.
                   If it is specified, then the session in particular opens
                   that db and will use it afterwards whether another one
                   was already opened or not.
        :param modules: any iterable of module names.
                        Not installed modules will be ignored
                        The special name ``'all'`` triggers the update of
                        all installed modules.
        """
        if db is None:
            if self.cr is None:
                raise ValueError("update_modules needs either the session to "
                                 "be opened or an explicit database name")
            db = self.cr.dbname

        if self.cr is not None:
            self.close()
        for module in modules:
            config['update'][module] = 1
        self._registry = openerp.modules.registry.RegistryManager.get(
            db, update_module=True)
        config['update'].clear()
        self.init_cursor()

    def install_modules(self, modules, db=None, update_modules_list=True,
                        open_with_demo=False):
        """Install the modules in the database.

        Has the side effect of closing the current cursor, committing if and
        only if the list of modules is updated.

        Demo data loading is handled consistently with the decision taken
        by :meth:`open`.

        :param db: Database name. If not specified, it is assumed to have
                   already been opened with :meth:`open`, e.g, for a prior
                   read of :meth:`db_version`.
                   If it is specified, then the session in particular opens
                   that db and will use it afterwards whether another one
                   was already opened or not.
        :param modules: any iterable of module names.
        :param update_modules_list: if True, will update the module lists
                                    *and commit* before the install begins.
        :param open_with_demo: if ``db`` is not None, will be passed to
                               :meth:`open`.
        """
        already_open = self.cr is not None
        if db is None:
            if not already_open:
                raise ValueError("install_modules needs either the session to "
                                 "be opened or an explicit database name")
            db = self.cr.dbname
        elif update_modules_list and not (
                already_open and self.cr.dbname == db):
            self.open(db=db, with_demo=open_with_demo)

        if update_modules_list:
            self.update_modules_list()
            self.cr.commit()

        if self.cr is not None:
            self.close()
        saved_without_demo = config['without_demo']

        # with update_modules_list=False, an explicitely named DB would not
        # have gone through open() yet.
        config['without_demo'] = not getattr(self, 'with_demo', open_with_demo)
        for module in modules:
            config['init'][module] = 1
        self._registry = openerp.modules.registry.RegistryManager.get(
            db, update_module=True, force_demo=self.with_demo)
        config['init'].clear()
        config['without_demo'] = saved_without_demo
        self.init_cursor()

    def handle_command_line_options(self, to_handle):
        """Handle prescribed command line options and eat them.

        Anything before first occurrence of ``--`` on the command-line is taken
        into account and removed from ``sys.argv``.

        Help messages:

        If -h or --help is specified  and -- is not, the help for the wrapper
        will be printed, and the -h/--help option kept in sys.argv.

        If -h or --help is specified before --, the help for this wrapper will
        be printed and options after -- will be kept in sys.argv.

        if -h or --help is specified after --, it will be ignored at this
        stage, and kept in sys.argv (in most cases triggering help print for
        the wrapped script).
        """

        parser = OptionParser(
            usage="%(prog)s [OpenERP options] -- other arguments",
            description="This is a script rewrapped by OpenERP buildout "
                        "recipe to add OpenERP-related options on the command "
                        "line prior to other arguments.")

        if '-d' in to_handle:
            parser.add_option('-d', '--db-name',
                              help="Name of the database to work on. "
                              "If not specified, the database from "
                              "configuration files will be used")

        try:
            sep = sys.argv.index('--')
        except ValueError:
            if '-h' in sys.argv or '--help' in sys.argv:
                # in case of call myscript -h --, only the wrapper help
                # will be printed
                parser.epilog = ("Help message from the wrapped script, "
                                 "if any, will follow.")
                parser.print_help()
                print
                return
            our_argv = []
            sep = None
        else:
            our_argv = sys.argv[1:sep]

        options, args = parser.parse_args(our_argv)

        if sep is not None:
            del sys.argv[1:sep+1]

        if '-d' in to_handle:
            if options.db_name:
                logger.info("Opening database %r", options.db_name)
            else:
                logger.info("No database specified, using the one specified "
                            "in buildout configuration.")
            self.open(db=options.db_name)



import sys
import os


def insert_args(arguments):
    for i, a in enumerate(arguments):
        sys.argv.insert(i+1, a)


def main(starter, conf, version=()):
    if version >= (6, 0):
        log_level = 'test'
    else:
        log_level = 'info'
    arguments = ['-c', conf, '--log-level', log_level, '--stop-after-init']
    if version >= (7, 0):
        arguments.append('--test-enable')

    insert_args(arguments)

    if version == (5, 0):
        from . import patch_openerp_v5  # noqa

    os.chdir(os.path.split(starter)[0])
    glob = globals()
    glob['__name__'] = '__main__'
    glob['__file__'] = starter
    execfile(starter, globals())



import unittest
import tempfile
import shutil
import os

from ..utils import working_directory_keeper


class WorkingDirectoryTestCase(unittest.TestCase):

    def setUp(self):
        self.dirpath = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.dirpath)

    def test_ctx_manager(self):
        # warning: if this fails, then many other tests can be
        # affected, because many don't like being in /
        current = os.getcwd()
        with working_directory_keeper:
            os.chdir(self.dirpath)
            self.assertNotEqual(os.getcwd(), current)

        self.assertEqual(os.getcwd(), current)



"""VCS tests: Bazaar."""

import os
import subprocess
from ..testing import COMMIT_USER_FULL
from ..testing import VcsTestCase
from ..bzr import BzrBranch
from ..bzr import working_directory_keeper
from ..base import UpdateError
from ..base import CloneError


class BzrBaseTestCase(VcsTestCase):
    """Common utilities for Bazaard test cases."""

    def create_src(self):
        os.chdir(self.src_dir)
        subprocess.call(['bzr', 'init', 'src-branch'])
        self.src_repo = os.path.join(self.src_dir, 'src-branch')
        os.chdir(self.src_repo)
        subprocess.call(['bzr', 'whoami', '--branch', COMMIT_USER_FULL])
        f = open('tracked', 'w')
        f.write("first" + os.linesep)
        f.close()
        subprocess.call(['bzr', 'add'])
        subprocess.call(['bzr', 'commit', '-m', 'initial commit'])
        f = open('tracked', 'w')
        f.write("last" + os.linesep)
        f.close()
        subprocess.call(['bzr', 'commit', '-m', 'last version'])

    def assertRevision(self, branch, rev, first_line):
        """Assert that branch is at prescribed revision

        Double check with expected first line of 'tracked' file."""
        target_dir = branch.target_dir
        self.assertTrue(os.path.isdir(target_dir))
        f = open(os.path.join(target_dir, 'tracked'))
        lines = f.readlines()
        f.close()
        self.assertEquals(lines[0].strip(), first_line)
        self.assertEquals(branch.parents(), [rev])

    def assertRevision1(self, branch):
        """Assert that branch is at revision 1."""
        self.assertRevision(branch, '1', 'first')

    def assertRevision2(self, branch):
        """Assert that branch is at revision 2."""
        self.assertRevision(branch, '2', 'last')


class BzrTestCase(BzrBaseTestCase):

    def test_branch(self):
        target_dir = os.path.join(self.dst_dir, "My branch")
        branch = BzrBranch(target_dir, self.src_repo)
        branch('last:1')
        self.assertRevision2(branch)

    def test_branch_on_revision_retry(self):
        """Test retry system if direct branching to revison fails."""
        target_dir = os.path.join(self.dst_dir, "My branch")
        branch = BzrBranch(target_dir, self.src_repo)

        normal_method = branch._branch
        monkey_called = []

        def branch_no_rev(revision):
            """Monkey patch to simulate the error."""
            monkey_called.append(revision)
            if revision:
                raise CloneError("fake branch cmd", 3)
            return normal_method(revision)

        branch._branch = branch_no_rev
        branch('last:1')

        # ensures that we actually did test something:
        self.assertEqual(monkey_called, ['last:1', None])
        self.assertRevision2(branch)  # branching worked

    def test_branch_options_conflict(self):
        target_dir = os.path.join(self.dst_dir, "My branch")
        branch = BzrBranch(target_dir, self.src_repo,
                           **{'bzr-init': 'branch',
                              'bzr-stacked-branches': 'True'})
        self.assertRaises(Exception, branch, "last:1")

    def test_branch_bzr_init(self):
        target_dir = os.path.join(self.dst_dir, "My branch")
        branch = BzrBranch(target_dir, self.src_repo,
                           **{'bzr-init': 'branch'})
        branch('last:1')
        self.assertRevision2(branch)

    def test_branch_stacked_deprecated(self):
        target_dir = os.path.join(self.dst_dir, "My branch")
        branch = BzrBranch(target_dir, self.src_repo,
                           **{'bzr-stacked-branches': 'True'})
        branch('last:1')
        self.assertRevision2(branch)

    def test_branch_stacked(self):
        target_dir = os.path.join(self.dst_dir, "My branch")
        branch = BzrBranch(target_dir, self.src_repo,
                           **{'bzr-init': 'stacked-branch'})
        branch('last:1')
        self.assertRevision2(branch)

    def test_checkout_lightweight(self):
        target_dir = os.path.join(self.dst_dir, "My branch")
        branch = BzrBranch(target_dir, self.src_repo,
                           **{'bzr-init': 'lightweight-checkout'})
        branch('1')
        self.assertRevision1(branch)

        branch('last:1')
        self.assertRevision2(branch)

    def test_branch_to_rev(self):
        """Directly clone and update to given revision."""
        target_dir = os.path.join(self.dst_dir, "My branch")
        branch = BzrBranch(target_dir, self.src_repo)
        branch('1')
        self.assertRevision1(branch)

    def test_update(self):
        """Update to a revision that's not the latest available in target"""
        target_dir = os.path.join(self.dst_dir, "clone to update")
        branch = BzrBranch(target_dir, self.src_repo)('last:1')

        # Testing starts here
        branch = BzrBranch(target_dir, self.src_repo)
        branch('1')
        self.assertRevision1(branch)

    def test_update_tag(self):
        """Update to an avalailable rev, identified by tag.
        """
        with working_directory_keeper:
            os.chdir(self.src_repo)
            subprocess.check_call(['bzr', 'tag', '-r', '1', 'sometag'])

        target_dir = os.path.join(self.dst_dir, "clone to update")

        # Testing starts here
        branch = BzrBranch(target_dir, self.src_repo)
        branch('sometag')
        self.assertRevision1(branch)

    def test_update_needs_pull(self):
        """Update to a revision that needs to be pulled from target."""
        target_dir = os.path.join(self.dst_dir, "clone to update")
        branch = BzrBranch(target_dir, self.src_repo)('1')
        # We really don't have rev 2 in branch
        self.assertRaises(LookupError, branch.get_revid, '2')

        branch('2')
        self.assertRevision2(branch)

    def test_update_revid_needs_pull(self):
        """Update to a rev that needs to be pulled from source, by revid."""
        target_dir = os.path.join(self.dst_dir, "clone to update")
        branch = BzrBranch(target_dir, self.src_repo)('1')
        # We really don't have rev 2 in branch
        self.assertRaises(LookupError, branch.get_revid, '2')

        revid = BzrBranch(self.src_repo, self.src_repo).get_revid('2')
        branch('revid:' + revid)
        self.assertRevision2(branch)

    def test_clean(self):
        """Test clean method, and esp. that no local mod occurs.

        See launchpad: #1192973, bzr tracks empty directories. We must not
        remove them.
        """
        with working_directory_keeper:
            os.chdir(self.src_repo)
            os.mkdir('subdir')
            subprocess.check_call(['bzr', 'add'])
            subprocess.check_call(['bzr', 'commit', '-m', "bidule"])
        target_dir = os.path.join(self.dst_dir, "clone to clean")
        branch = BzrBranch(target_dir, self.src_repo)
        try:
            branch.clean()
        except:
            self.fail("clean() should not fail if "
                      "branch not already done")

        branch('last:1')
        untracked = os.path.join(branch.target_dir, 'subdir', 'untracked.pyc')
        with open(untracked, 'w') as f:
            f.write('untracked content')
        branch.clean()
        self.assertFalse(os.path.exists(untracked))
        with working_directory_keeper:
            os.chdir(branch.target_dir)
            bzr_status = subprocess.Popen(['bzr', 'status'],
                                          stdout=subprocess.PIPE)
            out = bzr_status.communicate()[0]

        self.assertEquals(bzr_status.returncode, 0)
        # output of 'bzr status' should be empty : neither unknown file nor
        # any local modification, including removal of 'subdir'
        self.assertEquals(out.strip(), '')

    def test_uncommitted_changes_tracked(self):
        target_dir = os.path.join(self.dst_dir, "clone to dirty")
        branch = BzrBranch(target_dir, self.src_repo)('last:1')
        self.assertFalse(branch.uncommitted_changes())
        with open(os.path.join(target_dir, 'tracked'), 'w') as f:
            f.write('some change')
        self.assertTrue(branch.uncommitted_changes())

    def test_uncommitted_changes_untracked(self):
        target_dir = os.path.join(self.dst_dir, "clone to dirty")
        branch = BzrBranch(target_dir, self.src_repo)('last:1')
        self.assertFalse(branch.uncommitted_changes())
        with open(os.path.join(target_dir, 'unknownfile'), 'w') as f:
            f.write('some change')
        self.assertTrue(branch.uncommitted_changes())

    def test_archive(self):
        target_dir = os.path.join(self.dst_dir, "clone to archive")
        branch = BzrBranch(target_dir, self.src_repo)
        branch('1')

        archive_dir = os.path.join(self.dst_dir, "archive directory")
        branch.archive(archive_dir)
        with open(os.path.join(archive_dir, 'tracked')) as f:
            self.assertEquals(f.readlines()[0].strip(), 'first')

    def test_url_update(self):
        """Method to update branch.conf does it and stores old values"""
        # Setting up a prior branch
        target_dir = os.path.join(self.dst_dir, "clone to update")
        branch = BzrBranch(target_dir, self.src_repo)
        branch('1')
        # src may have become relative, let's keep it in that form
        old_src = branch.parse_conf()['parent_location']

        # first rename.
        # We test that pull actually works rather than
        # just checking branch.conf to avoid logical loop testing nothing
        new_src = os.path.join(self.src_dir, 'new-src-repo')
        os.rename(self.src_repo, new_src)
        branch = BzrBranch(target_dir, new_src)
        branch('last:1')

        self.assertEquals(branch.parse_conf(), dict(
            buildout_save_parent_location_1=old_src,
            parent_location=new_src))

        # second rename
        new_src2 = os.path.join(self.src_dir, 'new-src-repo2')
        os.rename(new_src, new_src2)
        branch = BzrBranch(target_dir, new_src2)
        branch('1')

        self.assertEquals(branch.parse_conf(), dict(
            buildout_save_parent_location_1=old_src,
            buildout_save_parent_location_2=new_src,
            parent_location=new_src2))

    def test_url_update_1133248(self):
        """Method to update branch.conf is resilient wrt to actual content.

        See lp:1133248 for details
        """
        # Setting up a prior branch
        target_dir = os.path.join(self.dst_dir, "clone to update")
        branch = BzrBranch(target_dir, self.src_repo)
        branch('1')

        conf_path = os.path.join(target_dir, '.bzr', 'branch', 'branch.conf')
        with open(conf_path, 'a') as conf:
            conf.seek(0, os.SEEK_END)
            conf.write(os.linesep + "Some other stuff" + os.linesep)

        # src may have become relative, let's keep it in that form
        old_src = branch.parse_conf()['parent_location']

        # first rename.
        # We test that pull actually works rather than
        # just checking branch.conf to avoid logical loop testing nothing
        new_src = os.path.join(self.src_dir, 'new-src-repo')
        os.rename(self.src_repo, new_src)
        branch = BzrBranch(target_dir, new_src)
        branch('last:1')

        self.assertEquals(branch.parse_conf(), dict(
            buildout_save_parent_location_1=old_src,
            parent_location=new_src))

    def test_lp_url(self):
        """lp: locations are being rewritten to the actual target."""
        branch = BzrBranch('', 'lp:anybox.recipe.openerp')
        # just testing for now that it's been rewritten
        self.failIf(branch.url.startswith('lp:'))

        # checking idempotency of rewritting
        branch2 = BzrBranch('', branch.url)
        self.assertEquals(branch2.url, branch.url)

    def test_lp_url_nobzrlib(self):
        """We can't safely handle lp: locations without bzrlib."""
        from anybox.recipe.openerp import vcs
        save = vcs.bzr.LPDIR
        vcs.bzr.LPDIR = None
        self.assertRaises(RuntimeError, BzrBranch, '', 'lp:something')
        vcs.bzr.LPDIR = save

    def test_update_clear_locks(self):
        """Testing update with clear locks option."""
        # Setting up a prior branch
        target_dir = os.path.join(self.dst_dir, "clone to update")
        BzrBranch(target_dir, self.src_repo)('last:1')

        # Testing starts here
        branch = BzrBranch(target_dir, self.src_repo, clear_locks=True)
        branch('1')
        self.assertRevision1(branch)

    def test_failed(self):
        target_dir = os.path.join(self.dst_dir, "My branch")
        branch = BzrBranch(target_dir, '/does-not-exist')
        self.assertRaises(subprocess.CalledProcessError,
                          branch.get_update, 'default')


class BzrOfflineTestCase(BzrBaseTestCase):

    def make_local_branch(self, path, initial_rev, options=None):
        """Make a local branch of the source at initial_rev and forbid pulls.
        """
        if options is None:
            options = {}
        target_dir = os.path.join(self.dst_dir, path)
        # initial branching (non offline
        BzrBranch(target_dir, self.src_repo)(initial_rev)

        # crippled offline branch
        branch = BzrBranch(target_dir, self.src_repo, offline=True, **options)

        def _pull():
            self.fail("Should not pull !")

        branch._pull = _pull
        return branch

    def test_update_needs_pull(self):
        """[offline mode] updating to a non available rev raises UpdateError.
        """
        branch = self.make_local_branch("clone to update", '1')
        self.assertRaises(UpdateError, branch, '2')

    def test_update_last(self):
        """[offline mode] update to a last:1 rev does nothing."""
        branch = self.make_local_branch("clone to update", '1')
        branch('last:1')
        self.assertRevision1(branch)

    def test_update_available_revno(self):
        """[offline mode] update to an available revno works"""
        branch = self.make_local_branch("clone to update", 'last:1')
        branch('1')
        self.assertRevision1(branch)

    def test_update_available_revid(self):
        """[offline mode] update to an available revid works.
        """
        branch = self.make_local_branch("clone to update", 'last:1')
        revid = branch.get_revid('1')
        branch('revid:' + revid)
        self.assertRevision1(branch)

    def test_lightweight_checkout_noupdate(self):
        """[offline mode] lightweight checkouts shall not be updated."""
        branch = self.make_local_branch(
            "clone to update", '1',
            options={'bzr-init': 'lightweight-checkout'})

        def _update(*a, **kw):
            self.fail("Should not update !")

        branch._update = _update

        branch('last:1')
        self.assertRevision1(branch)

    def test_lightweight_checkout_noupdate_fixed_rev(self):
        """[offline mode] lightweight checkouts shall not be updated."""
        branch = self.make_local_branch(
            "clone to update", 'last:1',
            options={'bzr-init': 'lightweight-checkout'})

        def _update(*a, **kw):
            self.fail("Should not update !")

        branch._update = _update

        branch('1')
        self.assertRevision2(branch)



"""This module bridges the classical ``openerp-server`` as a console script.

The :func:`main` function gets registered on the fly by the server recipe as
a console script entry point and used in particular for ``start_openerp`` and
``test_openerp``.

Some version independence logic for the startup process also get bootstrapped
from here.
"""

import sys
import os
from . import patch_openerp_v5


def insert_args(arguments):
    """Insert `arguments` in ``sys.argv`` (direct impact on child script).
    """
    for i, a in enumerate(arguments):
        sys.argv.insert(i+1, a)


def main(starter, conf, version=None, just_test=False):
    """Call the `starter` script, dispatching configuration.

    All arguments are set in the standalone script produced by buildout through
    entry point options.

    :param starter: path to the main script source file (currently
      ``openerp-server``)
    :param conf: path to the OpenERP configuration file (managed by the recipe)
    :param version: OpenERP major version
    :type version: tuple of integers
    :param just_test: if True, only run unit tests
    """
    arguments = ['-c', conf]

    if just_test:
        arguments.extend(('--log-level',
                          'test' if version >= (6, 0) else 'info',
                          '--stop-after-init'))

        if version >= (7, 0):
            arguments.append('--test-enable')

    if '--install-all' in sys.argv:
        sys.argv.remove('--install-all')
        from openerp.tools import config
        # Maybe we should preparse config in all cases and therefore avoid
        # adding the '-c' on the fly ?
        # Still, cautious about pre-6.1 versions
        config.parse_config(['-c', conf])
        from openerp.modules import get_modules
        arguments.extend(('-i', ','.join(get_modules())))

    insert_args(arguments)

    if version == (5, 0):
        patch_openerp_v5.do_patch()

    os.chdir(os.path.split(starter)[0])
    glob = globals()
    glob['__name__'] = '__main__'
    glob['__file__'] = starter
    try:
        execfile(starter, globals())
    except SystemExit as exc:
        if version == (5, 0):
            # Without Agent.quit() the Timer threads may go on forever
            # and the script will never stop
            import netsvc
            netsvc.Agent.quit()
        return exc.code



"""
Runtime
=======

This subpackage provides encapsulations and entry points for the application
itself:

* the ``session`` module features the supporting objects for "OpenERP scripts"
  and the dedicated python interpreter.
* the ``start_openerp`` and ``test_openerp`` modules are the entry points for
  the main startup scripts.

This architecture is meant in particular to provide stability and uniformity
accross OpenERP major versions, so that the recipe can be leveraged by
automated deploymnent tools and continuous integration systems.
"""

_imported_addons = set()


def already_imported(module_name):
    """Convenience to help some OpenERP modules to avoid been imported twice.

    Each call of this function returns a boolean indicated whether the
    specified module was already in the ``imported_addons`` registry and add it
    inconditionnally.

    Thus caller code is expected to import the module right away if the
    return value was False.
    """
    name = module_name.rsplit('.', 1)[-1]
    if name in _imported_addons:
        return True
    _imported_addons.add(name)
    return False


def clear_import_registry():
    _imported_addons.clear()
